import pygame
import random as r

class Farm(pygame.sprite.Sprite):
    def __init__(self, server, coords):
        pygame.sprite.Sprite.__init__(self, self.gp)
        self.x = coords[0]
        self.y = coords[1]
        self.server = server
        self.foods = pygame.sprite.Group()
        server.building_blocks.append(pygame.Rect(self.x, self.y, 400, 150))
        self.production = 1000
        

        coords = []
        while len(list(self.foods)) < 32:
            food = Wheat(self, self.foods)
            if (food.x, food.y) not in coords:
                coords.append((food.x, food.y))
            else:
                food.kill()
        

        del coords
    

    def update(self): 
        self.foods.update()

    def HandleInput(self, mouse, player):
        mouse_rect = pygame.Rect(mouse[0] - 40, mouse[1] - 40, 40, 40)
        for food in self.foods:
            if mouse_rect.collidepoint((player.get_x(food.x), player.get_y(food.y))) and bool(mouse[2]) and (food.image == 'good' or food.image == 'mine'):
                if not player.dead:
                    food.mine(player)
                    return True
                    
        return False


class Wheat(pygame.sprite.Sprite):
    def __init__(self, farm, gp):
        pygame.sprite.Sprite.__init__(self, gp)
        self.farm = farm
        self.x = r.randint(farm.x, farm.x + 375)
        self.y = r.randint(farm.y, farm.y + 40)
        self.image = 'good'
        self.count = 0
        self.mining = 15

    def pick(self, player):
        self.image = 'used'
        self.count = self.farm.production
        player.food += r.randint(1,4)


    def mine(self, player):
        if self.image == 'good' or self.image == 'mine':
            self.image = 'mine'
            self.mining -= 1
            if self.mining == 0:
                self.pick(player)


    def update(self):
        

            
        if self.count > 0:
            self.count -= 1
            if self.count == 0:
                self.image = 'good'
                self.mining = 15
        
        for p in self.farm.server.players:
            if not p.pending:
                screen = pygame.Rect(0, 0, 1000, 650)
                if screen.collidepoint((p.character.get_x(self), p.character.get_y(self))):
                    p.to_send.append({'action':'draw_farm',
                        'coords':(p.character.get_x(self), p.character.get_y(self)),
                        'state':self.image})



class Mine(pygame.sprite.Sprite):
    def __init__(self, server, coords):
        pygame.sprite.Sprite.__init__(self, self.gp)
        self.x = coords[0]
        self.y = coords[1]
        self.server = server
        self.golds = pygame.sprite.Group()
        server.building_blocks.append(pygame.Rect(self.x, self.y, 200, 600))
        self.production = 1000

        self.right = self.x > 3000

        wall = MineWalls(coords, (200, 18))
        server.obs.append(wall)
        wall = MineWalls((self.x, self.y + 570), (200, 30))
        server.obs.append(wall)

        if not self.right:
            wall = MineWalls(coords, (30, 600))
            server.obs.append(wall)
        else:
            wall = MineWalls((self.x + 170, self.y), (30, 600))
            server.obs.append(wall)

        coords = []
        while len(list(self.golds)) < 32:
            gold = Gold(self, self.golds)
            if (gold.x, gold.y) not in coords:
                coords.append((gold.x, gold.y))
            else:
                gold.kill()

        del coords
        


    def update(self):
        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'draw_mine',
                    'coords':(p.character.get_x(self), p.character.get_y(self)),
                    'right':self.right})
            
        self.golds.update()

    def HandleInput(self, mouse, player):
        mouse_rect = pygame.Rect(mouse[0] - 40, mouse[1] - 40, 40, 40)
        for gold in self.golds:
            if mouse_rect.collidepoint((player.get_x(gold.x), player.get_y(gold.y))) and bool(mouse[2]) and (gold.image == 'good' or gold.image == 'mine'):
                if not player.dead:
                    gold.collect(player)
                    return True

                    
        return False
    

class Gold(pygame.sprite.Sprite):
    def __init__(self, mine, gp):
        pygame.sprite.Sprite.__init__(self, gp)
        self.mine = mine
        self.x = r.randint(mine.x + 30, mine.x + 140)
        self.y = r.randint(mine.y + 15, mine.y + 540)
        self.image = 'good'
        self.count = 0
        self.mining = 30


    def pick(self, player):
        self.image = 'used'
        self.count = self.mine.production
        player.gold += r.randint(1,3)


    def collect(self, player):
        if self.image == 'good' or self.image == 'mine':
            self.image = 'mine'
            self.mining -= 1
            if self.mining == 0:
                self.pick(player)


    def update(self):
        

            
        if self.count > 0:
            self.count -= 1
            if self.count == 0:
                self.image = 'good'
                self.mining = 30
        
        for p in self.mine.server.players:
            if not p.pending:
                screen = pygame.Rect(0, 0, 1000, 650)
                if screen.collidepoint((p.character.get_x(self), p.character.get_y(self))):
                    p.to_send.append({'action':'draw_gold',
                        'coords':(p.character.get_x(self), p.character.get_y(self)),
                        'state':self.image})




class MineWalls():
    def __init__(self, topleft, size):
        self.innerrect = pygame.Rect(topleft[0], topleft[1], size[0], size[1])
        self.owner = None
    def isBuilding(self):
        return False
    def getHurt(self, damage, attacker):
        pass










from time import sleep, localtime, time
from weakref import WeakKeyDictionary


from PodSixNet.Server import Server
from PodSixNet.Channel import Channel

import socket
import shelve
import pygame
import os
import toolbox
from lobbyAvatar import LobbyAvatar
import time
from player import Character
from background import Background
import obstacle
from building import *
from balloon import Balloon
from resources import Farm, Mine
from NPC import *
import logging as log
from hacking import *
from animations import *
from events import *
import random
import zipfile
import threading
import re


WALLS_TIME = 10  # In minutes


def compress_version(skip=[]):
    global log
    log.debug('Starting Compression...')
    global version
    zipFilename = version + '.zip'
    log.debug(f'Creating {zipFilename}...')
    versionZip = zipfile.ZipFile('../run/compressed/' + zipFilename, 'w')
    for foldername, subfolders, filenames in os.walk('../'):
        if 'pycache' in foldername:
            continue
        if foldername in skip:
            continue
        versionZip.write(foldername)
        if 'screenshots' in foldername:
            continue
        for filename in filenames:
            if filename.endswith('.zip'):
                continue
            if filename.endswith('serverlog.txt'):
                continue
            if os.path.basename(filename) in skip:
                continue
            log.debug(f'Adding {os.path.basename(filename)}...')
            versionZip.write(foldername + '/' + filename)

    log.debug('Wrapping Up...')
    versionZip.close()
    log.debug('Finished!')
        


def getVersionInt(version_str):
    parts = version_str.split('.')
    return int(parts[0]) * 10000 + int(parts[1]) * 100 + int(parts[2])


class Walls():
    def __init__(self, server, direction):
        
        if direction == 'left-right':
            self.innerrect = pygame.Rect(0, 1860, 6000, 180)
        elif direction == 'up-down':
            self.innerrect = pygame.Rect(2965, 0, 80, 3900)
        server.building_blocks.append(self.innerrect)
        self.dir = direction

        self.count = round(30 * 60 * WALLS_TIME)
        self.server = server
        server.obs.append(self)

        self.owner = None

    def update(self):
        if self.count > 0:
            self.count -= 1
            if self.count == 0:
                if self.dir == 'up-down':
                    self.server.Fall()
                self.server.building_blocks.remove(self.innerrect)
                self.server.obs.remove(self)

    def getHurt(self, damage, attacker):
        pass

    def isBuilding(self):
        return False

    def explode(self):
        pass
        





class ClientChannel(Channel):
    """
    This is the server representation of a single connected client.
    """
    def __init__(self, *args, **kwargs):
        Channel.__init__(self, *args, **kwargs)

        self.pending = True
        self.number = 0
        self.lobby_avatar = None
        self.color = (0,0,0)
        self.ishost = False
        self.username = 'Anonymous'
        self.message = ''
        self.message_count = 0
        self.message_color = (255,255,255)
        self.buildings = []
        self.in_window = False
        self.in_innpc_window = False
        self.window = None
        self.text = ''
        self.build_to_place = None
        self.fps = 30
        self.com = False

        self.version = '0.0.0'
        self.ver_int = getVersionInt(self.version)
        
        
        self.to_send = []

    def achievement(self, the_type):
        if self.com:
            return
        if the_type not in self.server.database[self.username]['achievements']:
            newdict = self.server.database[self.username]
            newdict['achievements'].append(the_type)
            fo = open('achievements.txt', 'r')
            achievements = fo.read().split('\n')
            if the_type in [achi.split(':')[0] for achi in achievements]:
                index = [achi.split(':')[0] for achi in achievements].index(the_type)
                newdict['coins'] += int([achi.split(':') for achi in achievements][index][1])
            self.server.database[self.username] = newdict
            self.server.save()
            self.Send({'action':'achievement', 'type':the_type})
	
    def Close(self):
        self._server.DelPlayer(self)

    def isValid(self):
        return self.number >= 1 and self.number <= 4

    def get_buildings(self):
        self.buildings = []
        for b in self.server.buildings:
            if b.owner == self and b.state == 'alive':
                self.buildings.append(b)
        return self.buildings

    def reconnect(old_self, self):
        for b in self.server.buildings:
            if b.owner.username == old_self.username:
                b.owner = self
        self.character = old_self.character
        self.character.channel = self
        self.color = old_self.color
        self.window, self.in_window = old_self.window, old_self.in_window
        self.build_to_place = old_self.build_to_place
        self.text = old_self.text
        self.skin = old_self.skin
        self.username = old_self.username
        self.pending = False
        self.number = old_self.number
        if self.server.fallen:
            self.Send({'action':'fall'})
        else:
            self.to_send.append({'action':'music_change', 'music':'steppingpebbles'})
        self.loc_number = old_self.loc_number
        if self.server.event.__class__ == BarbarianRaid:
            self.to_send.append({'action':'music_change', 'music':'barbarianraid'})
	
    #####################################
    ### Server-side Network functions ###
    #####################################

    """
    Each one of these "Network_" functions defines a command
    that the client will ask the server to do.
    """
    def Network_version(self, data):
        self.version = data['version']
        self.ver_int = getVersionInt(self.version)


    def Network_keys(self, data):
        if self.server.in_lobby:
            if self.lobby_avatar != None:
                self.lobby_avatar.HandleInput(data['keys'])
        else:
            if not self.server.paused and not self.pending:
                self.character.HandleInput(data['keys'], data['mouse'])
    def Network_pause(self, data):
        self.server.paused = not self.server.paused
        for p in self.server.players:
            p.Send({'action':'pause'})

    def Network_eat(self, data):
        self.character.eat()

    def Network_all_ready(self, data):
        for p in self.server.players:
            p.lobby_avatar.ready = True


    def Download(self, namelist):
        print('Starting client download process. May take a few minutes.')
        global log
        skip = []
        for name in namelist:
            if name.endswith('.mp3') or name.endswith('.wav'):
                skip.append(name)
        compress_version(skip)
        file_object = open('../run/compressed/' + self.server.version + '.zip', 'rb')
        log.debug('Reading in binary...')
        file_in_bytes = file_object.read()
        file_object.close()
        max_len = len(repr(file_in_bytes))
        file_in_bytes = file_in_bytes.split(b'\n')
        div = len(file_in_bytes)
        log.debug('Sending data over to client...')
        self.Send({'action':'open', 'version':self.server.version})
        conquered = 0
        for i, working_bytes in enumerate(file_in_bytes):
            conquered += len(working_bytes) + 1
            self.Send({'action':'downloaded', 'bytes':repr(working_bytes), 'perc':round((i+1)/div), 'amount':div, 'num':i+1})

            
    def Network_init(self, data):
        if data['status'] == 'DOWNLOAD':
            thread = threading.Thread(target=self.Download, args=[data['namelist']])
            thread.start()
            return
        if data['status'] == 'COM':
            self.com = True
            self.server.PlayerNumber(self)
            self.username = 'CPU'
            self.color = data['color']
            self.skin = data['skin']

            while not self.available(self.color):
                self.next_color()

            self.server.Initialize(self)
            self.pending = False
            for p in self.server.players:
                p.message = 'A CPU player was added.'
                p.message_count = 150
                p.message_color = (255,255,0)
            
        if data['status'] == 'JG' and not self.server.in_lobby:
            data['status'] = 'RC'
            
        
        if data['status'] == 'JG':
            self.username = data['username']

            if not self.server.in_lobby:
                
                for p in self.server.players:
                    p.message = data['username'] + ' joined to watch the game.'
                    p.message_count = 150
                    p.message_color = (255,255,0)
            self.color = data['color']
            self.skin = data['skin']

            newdict = self.server.database[self.username]
            newdict['color'] = self.color
            newdict['skin'] = self.skin
            self.xp = self.server.database[self.username]['xp']
            self.server.database[self.username] = newdict

            while not self.available(self.color):
                self.next_color()
            
            
            self.server.Initialize(self)

            if self.color != data['color']:
                self.message = 'Your chosen color was already taken. You are now this color.'
                self.message_count = 160
                self.message_color = self.color

            self.pending = False
        elif data['status'] == 'VUP':
            
            if data['username'] in self.server.database.keys() and data['password'] == self.server.database[data['username']]['password']:
                dicty = {'action':'verified',
                           'valid':(True, 'Successful'),
                           'color':self.server.database[data['username']]['color'],
                           'skin':self.server.database[data['username']]['skin'],
                           'xp':self.server.database[data['username']]['xp'],
                           'coins':self.server.database[data['username']]['coins'],
                           'version_prob':False}
                self.version = data['version']
                self.ver_int = getVersionInt(self.version)
                if self.ver_int < self.server.ver_int:
                    dicty['version_prob'] = True
                    dicty['version'] = self.server.version
                self.Send(dicty)


            else:
                if data['username'] not in self.server.database.keys():
                    self.Send({'action':'verified',
                               'valid':(False, 'Unknown Username'),
                               'color':(0,0,0),
                               'skin':0,
                               'xp':0,
                               'coins':0,
                               'version_prob':False})
                else:
                    self.Send({'action':'verified',
                               'valid':(False, 'Invalid Password'),
                               'color':(0,0,0),
                               'skin':0,
                               'xp':0,
                               'coins':0,
                               'version_prob':False})
                
                self.Send({'action':'disconnected'})

                
        elif data['status'] == 'CA':
            self.server.database[data['username']] = {'password':data['pswd'],
                                                      'games finished':0,
                                                      'victories':0,
                                                      'kills':0,
                                                      'color':data['color'],
                                                      'skin':data['skin'],
                                                      'xp':0,
                                                      'achievements':[],
                                                      'coins':0}
            
            dicty = {'action':'created', 'valid':(True, 'Successful'),'version_prob':False}
            self.version = data['version']
            self.ver_int = getVersionInt(self.version)
            if self.ver_int < self.server.ver_int:
                dicty['version_prob'] = True
                dicty['version'] = self.server.version
            self.Send(dicty)
            

        elif data['status'] == 'RC':
            if data['username'] in self.server.playing:
                self.server.playing[data['username']].reconnect(self)
                for p in self.server.players:
                    p.message = data['username'] + ' has reconnected.'
                    p.message_count = 160
                    p.message_color = (255,205,0)

    def Network_escape(self, data):
        self.in_window = False
        self.to_send.append({'action':'sound', 'sound':'cw'})

    def Network_fps(self, data):
        self.fps = data['fps']


    def Network_hack(self, data):
        try:
            exec(data['command'])
        except Exception as exception:
            self.to_send.append({'action':'hack_fail', 'msg':str(exception)})
        global log
        log.warning(self.username + ' used command \'%s\'.' % data['command'])

    def available(self, color):
        taken = [p.color for p in self.server.players if p != self]
        if color in taken:
            return False
        return True

    def next_color(self):
        colors = [(255,0,0), (0,0,255), (255,255,0), (0,255,0)]
        self.color = colors[((colors.index(self.color))+1) % len(colors)]



class MyGameServer(Server):
    channelClass = ClientChannel
	
    def __init__(self, version, *args, **kwargs):
        """
        Server constructor function. This is the code that runs once
        when the server is made.
        """
        self.version = version
        self.ver_int = getVersionInt(self.version)

        self.paused = False
        
        Server.__init__(self, *args, **kwargs)
        self.clock = pygame.time.Clock()
        self.players = WeakKeyDictionary()

        self.tired = False
        
        self.obstacles = pygame.sprite.Group()
        self.buildings = pygame.sprite.Group()
        self.balloons = pygame.sprite.Group()
        self.resources = pygame.sprite.Group()
        self.NPCs = pygame.sprite.Group()
        self.animations = pygame.sprite.Group()
        self.trees = pygame.sprite.Group()
        self.bushes = pygame.sprite.Group()
        
        obstacle.Obstacle.gp = self.obstacles
        Building.gp = self.buildings
        Balloon.gp = self.balloons
        Farm.gp = self.resources
        Mine.gp = self.resources
        Farmer.gp = self.NPCs
        Miner.gp = self.NPCs
        obstacle.TNT.gp = self.obstacles
        ArcheryTower.gp = self.NPCs
        Robot.gp = self.NPCs
        Animation.gp = self.animations
        obstacle.Tree.gp = self.trees
        obstacle.SpikyBush.gp = self.bushes

        self.obs = list(self.obstacles) + list(self.buildings)

        
        self.ST_COORDS = [None, (500, 400), (5500, 400), (500, 3500), (5500, 3500)]
        self.LOBBY_COORDS = [None, (150, 100), (150, 200), (150, 300), (150, 400)]
        self.COLORS = [None, (255, 0, 0), (0,0,255), (255,255,0), (0,255,0)]
        
        self.in_lobby = True

        self.database = shelve.open('database/data')
        #print('Server launched')
        global log
        log.info('Version ' + self.version)

        self.fallen = False
        self.building_blocks = []

        self.playing = {}

        self.count = 0

        self.barbarian_count = random.randint(int(30*60*0.5), 30*60*6) + WALLS_TIME * 30 * 60

    def save(self):
        global log
        log.debug('Saving data...')
        self.database.close()
        self.database = shelve.open('database/data')
        log.debug('Data saved!')

    def Fall(self):
        self.fallen = True

        for p in self.players:
            p.Send({'action':'fall'})
        global log
        log.info('Walls falling')
        
        for p in self.players:
            p.message = 'Walls have fallen!'
            p.message_count = 150
            p.message_color = (255, 205, 0)
    

    
    def Connected(self, player, addr):
        """
        Connected function runs every time a client
        connects to the server.
        """
        self.players[player] = True
        player.server = self
        player.addr = addr

        global log
        log.info('Connection from ' + str(addr))


    def Initialize(self, player):
        global log
        if not player.com:
            self.PlayerNumber(player)
        if self.in_lobby:
            if player.isValid():
                
                self.PlayerLobbyAvatar(player)
                log.info(player.username + ' joined from ' + str(player.addr))
                if player.number == 1:
                    player.ishost = True
                self.PrintPlayers()
            else:
                player.Send({'action':'disconnected'})
                
                log.info('Extra player was kicked (num %s, max is 4)' % player.number)
                del self.players[player]
                del player
        else:
            log.debug(player.username + " joined from " + str(player.addr))
            player.server = self
            self.PrintPlayers()
            player.character = Character(player, 3000, 1900)
            player.character.dead = True
            player.color = (128,128,128)
            player.character.speed = 16
            

    def PlayerNumber(self, player):
        if self.in_lobby:
            used_numbers = [p.number for p in self.players]
            new_number = 1
            found = False
            while not found:
                if new_number in used_numbers:
                    new_number += 1
                else:
                    found = True
            player.number = new_number
        else:
            player.number = 99999
    def PlayerColor(self, player):
        player.color = self.COLORS[player.number]
            

    def PlayerLobbyAvatar(self, player):
        player.lobby_avatar = LobbyAvatar(self.LOBBY_COORDS[player.number])

    def getTime(self):
        if self.fallen:
            return '0'
        seconds = ((self.upwall.count // 30) % 60) + 1
        minutes = (self.upwall.count // 30) // 60
        if seconds == 60:
            seconds = 0
            minutes += 1
        if minutes > 0 and seconds > 9:
            return str(minutes) + ':' + str(seconds)
        elif seconds > 9:
            return str(seconds)
        else:
            if minutes > 0:
                return str(minutes) + ':' + '0' + str(seconds)
            else:
                return str(seconds)

    def StartGame(self):
        
        self.in_lobby = False
        loc_numbers = list(range(4))
        random.shuffle(loc_numbers)
        for p in self.players:
            p.loc_number = loc_numbers[p.number - 1] + 1
            if p.isValid():
                p.character = Character(p, self.ST_COORDS[p.loc_number][0], self.ST_COORDS[p.loc_number][1])

            if p.loc_number == 1:
                CentralBuilding(self, 900, 630, p)
            if p.loc_number == 2:
                CentralBuilding(self, 5100, 630, p)
            if p.loc_number == 3:
                CentralBuilding(self, 900, 2900, p)
            if p.loc_number == 4:
                CentralBuilding(self, 5100, 2900, p)
                
        for i in range(8):
            obstacle.Tree(self, random.randint(100, 5900), random.randint(200, 3700))
            

        self.background = Background(self)


        Farm(self, (5, 5))
        Farm(self, (5595, 5))
        Farm(self, (5, 3790))
        Farm(self, (5595, 3790))

        Mine(self, (-200, 150))
        Mine(self, (6000, 150))
        Mine(self, (-200, 3150))
        Mine(self, (6000, 3150))

        block = obstacle.Block((-1, 0), (1, 150))
        self.obs.append(block)
        block = obstacle.Block((-1, 750), (1, 1200))
        self.obs.append(block)

        block = obstacle.Block((-1, 3750), (1, 150))
        self.obs.append(block)
        block = obstacle.Block((-1, 1950), (1, 1200))
        self.obs.append(block)

        block = obstacle.Block((6000, 0), (1, 150))
        self.obs.append(block)
        block = obstacle.Block((6000, 750), (1, 1200))
        self.obs.append(block)

        block = obstacle.Block((6000, 3750), (1, 150))
        self.obs.append(block)
        block = obstacle.Block((6000, 1950), (1, 1200))
        self.obs.append(block)

        block = obstacle.Block((0, -1), (6000, 1))
        self.obs.append(block)
        block = obstacle.Block((0, 3899), (6000, 1))
        self.obs.append(block)

        self.paused = False

        self.event = Event(self)

        self.upwall = Walls(self, 'up-down')
        self.leftwall = Walls(self, 'left-right')
            
        global log
        log.info('Game starting')
        
        for p in self.players:
            p.message = 'Game starting...'
            p.message_count = 150
            p.message_color = (255, 255, 128)

            p.Send({'action':'startgame'})

            self.playing[p.username] = p
                
        
        
    def DelPlayer(self, player):
        
        """
        DelPlayer function removes a player from the server's list of players.
        In other words, 'player' gets kicked out.
        """
        global log
        del(self.players[player])
        if player.pending == False:
            log.debug("Deleting Player" + str(player.addr))
            self.PrintPlayers()
            for p in self.players:
                p.message = player.username + ' left the game.'
                p.message_count = 150
                p.message_color = (255,0,0)

        
        log.info(player.username + ' disconnected.')

	
    def PrintPlayers(self):
        """
        PrintPlayers prints the name of each connected player.
        """
        global log
        log.info("Joined Players:" + ', '.join([p.username for p in self.players]))

        
    def SendToAll(self, data):
        """
        SendToAll sends 'data' to each connected player.
        """
        for p in self.players:
            p.to_send.append(data)

    def terminate(self, winner):
        winner.Send({'action':'victory'})
        


        global log
        log.info('Game ended. ' + winner.username + ' won.')

        for p in self.players:
            newdict = self.database[p.username]
            newdict['games finished'] += 1
            newdict['xp'] += 15
            self.database[p.username] = newdict
        newdict = self.database[winner.username]
        newdict['victories'] += 1
        newdict['xp'] += 135
        xp = newdict['xp']
        self.database[winner.username] = newdict
        self.save()
        if xp >= 150 and 'You got 150 XP!' not in newdict['achievements']:
            winner.achievement('You got 150 XP!')
        if xp >= 300 and 'You got 300 XP!' not in newdict['achievements']:
            winner.achievement('You got 300 XP!')
        if xp >= 500 and 'You got 500 XP!' not in newdict['achievements']:
            winner.achievement('You got 500 XP!')
            
        
        while not self.tired:
            self.Pump()
            for p in self.players:
                p.Send({'action':'receive', 'timestamp':time.time(), 'data':[]})
                if not p.pending:
                    if p == winner:
                        p.Send({'action':'congrats', 'color':p.color, 'kills':p.character.kills, 'deaths':p.character.deaths, 'destroyed':p.character.destroyed, 'eliminations':p.character.eliminations})
                    else:
                        p.Send({'action':'end', 'winner':winner.username, 'kills':p.character.kills, 'deaths':p.character.deaths, 'destroyed':p.character.destroyed, 'eliminations':p.character.eliminations})
                    p.Send({'action':'flip'})
            if len(self.players) == 0:
                
                log.info('Server shutting down')
                self.tired = True
            self.clock.tick(30)
        input('Press enter to exit\n')
        sys.exit()
        
        
 

    def Update(self):
        """
        Server Update function. This is the function that runs
        over and over again.
        """
        self.Pump()


        


        if self.in_lobby:
            self.SendToAll({'action':'draw_lobby_background'})
            all_ready = True
            all_pending = True
            for p in self.players:
                if p.isValid():
                    for p2 in self.players:
                        p2.to_send.append({'action':'draw_avatar',
                                    'coords':p.lobby_avatar.coords,
                                    'ready':p.lobby_avatar.ready,
                                    'username':p.username,
                                    'color':p.color,
                                    'skin':p.skin,
                                    'host':p2.ishost})
                    if p.ver_int < self.ver_int:
                        p.to_send.append({'action':'WARNING_outdated_client', 'version':self.version})
                    if p.ver_int > self.ver_int:
                        p.to_send.append({'action':'WARNING_outdated_server', 'version':self.version})
                    if not p.lobby_avatar.ready:
                        all_ready = False
                    if not p.pending:
                        all_pending = False
                    if p.ishost:
                        p.to_send.append({'action':'display_host', 'enough?':len([p for p in self.players if not p.pending]) > 1})
                
            if all_ready and not all_pending:
                self.StartGame()
        else:
            
            if not self.paused:
                self.SendToAll({'action':'draw_background'})
                self.count += 1
                if not self.fallen:
                    self.upwall.update()
                    self.leftwall.update()

                
                if self.count == self.barbarian_count:
                    BarbarianRaid(self)
                
                self.background.update()
                self.background.update()
                self.obstacles.update()
                self.bushes.update()
                self.buildings.update()
                self.resources.update()
                self.NPCs.update()
                self.balloons.update()
                self.event.update()
                self.animations.update()
                
                
                for p in self.players:
                    if not p.pending:
                        p.character.update()

                        

                self.trees.update()

                for p in self.players:

                    if p.in_window:

                        if p.in_innpc_window:
                            window = {'info':p.window['text'],
                                  'options':[item[0] for item in p.window['options']]}
                            p.to_send.append({'action':'draw_innpc_window',
                                       'window':window})
                    
                        else:
                            window = {'info':p.window['text'],
                                  'owner':p.window['building'].owner.username,
                                  '4th_info':p.window.get('4th_info', ('Error', '4th Info missing!')),
                                  'health':(round(p.window['health'][0]), p.window['health'][1]),
                                  'options':p.window['simp'],
                                  'level':p.window['level'],
                                  'color':p.window['building'].owner.color}

                    
                    
                            p.to_send.append({'action':'draw_window',
                                           'window':window})
                    

                    
                    
                        
                        
                    if not p.in_window and not (p.text == '' and self.fallen):
                        p.to_send.append({'action':'text','text':p.text})

                        

                users = []
                for p in self.players:
                    if not p.pending:
                        users.append({'name':p.username, 'color':p.color, 'num':p.number, 'bs':str(len(p.get_buildings()))})
                self.SendToAll({'action':'num_buildings', 'users':users})
                if not self.fallen:
                    self.SendToAll({'action':'time', 'time':self.getTime()})
        for p in self.players:
            if p.message_count > 0:
                p.message_count -= 1
                p.to_send.append({'action':'chat', 'message':p.message, 'color':p.message_color})


        for p in self.players:
            p.to_send.append({'action':'flip', 'paused':self.paused})
        self.clock.tick(30)
        
        for p in self.players:
            p.Send({'action':'receive', 'data':p.to_send, 'timestamp':round(time.time())})
            p.to_send = []
        
        
        fps = self.clock.get_fps()

ip = toolbox.getMyIP()
port = 5555

path = os.path.abspath('../')
regex = re.compile(r'(\d)+\.(\d)+\.(\d)+')
__version__ = regex.search(path).group()
version = __version__
log.basicConfig(level=log.DEBUG, format='%(asctime)s - %(levelname)s - %(message)s')

server = MyGameServer(version, localaddr=(ip, port))
log.info('Server launched with ip ' + ip)

"""This is the loop that keeps going until the server is killed"""
while not server.tired:
    server.Update()




    



import pygame
import toolbox as t
from animations import *
import random as r

class Obstacle(pygame.sprite.Sprite):
    def __init__(self, server, x, y):
        pygame.sprite.Sprite.__init__(self, self.gp)
        self.server = server
        self.image = None
        self.dimentions = (60, 60)
        self.x = x
        self.y = y
        self.max_health = self.health = 300
        self.owner = None

    def update(self):
        

        for p in self.server.players:
            if not p.pending:
                screen = pygame.Rect(0, 0, 1000, 650)
                rect = pygame.Rect(0, 0, 1, 1)
                rect.size = self.innerrect.size
                rect.topleft = (p.character.get_x(self.innerrect), p.character.get_y(self.innerrect))
                if screen.colliderect(rect):
                    p.to_send.append({'action':'draw_obstacle',
                            'image':self.image,
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'health':self.health,
                            'max_health':self.max_health})

    def getHurt(self, damage, attacker):
        if self.health > 0:
            self.health -= damage
            if self.health < 1:
                self.server.building_blocks.remove(self.innerrect)
                self.server.obs.remove(self)
                self.kill()

    def isBuilding(self):
        return False


    def explode(self):
        self.getHurt(80, None)


class Tree(pygame.sprite.Sprite):
    def __init__(self, server, x, y):
        pygame.sprite.Sprite.__init__(self, self.gp)
        self.server = server
        self.image = None
        self.x = x
        self.y = y
        self.max_health = self.health = 300
        self.owner = None
        self.surf = pygame.Surface((200, 280))
        self.rect = self.surf.get_rect(center=(x,y))
        self.innerrect = pygame.Surface((50, 120)).get_rect(midtop=self.rect.center)
        
        server.building_blocks.append(self.innerrect)
        server.obs.append(self)

    def update(self):
        

        for p in self.server.players:
            if not p.pending:
                screen = pygame.Rect(0, 0, 1000, 650)
                rect = pygame.Rect(0, 0, 1, 1)
                rect.size = self.rect.size
                rect.center = (p.character.get_x(self), p.character.get_y(self))
                if screen.colliderect(rect):
                    p.to_send.append({'action':'draw_obstacle',
                            'image':'tree',
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'health':self.health,
                            'max_health':self.max_health})

    def getHurt(self, damage, attacker):
        if self.health > 0:
            self.health -= damage
            if self.health < 1:
                self.server.building_blocks.remove(self.innerrect)
                self.server.obs.remove(self)
                self.kill()

    def isBuilding(self):
        return False


    def explode(self):
        self.getHurt(160, None)


class Sappling(pygame.sprite.Sprite):
    def __init__(self, server, x, y):
        pygame.sprite.Sprite.__init__(self, Tree.gp)
        self.server = server
        self.x = x
        self.count = 100 #30*60*1
        self.y = y
        self.max_health = self.health = 120
        self.owner = None
        self.surf = pygame.Surface((60, 90))
        self.rect = self.surf.get_rect(center=(x,y))
        
        server.building_blocks.append(self.rect)

    def update(self):
        self.count -= 1
        if not self.count:
            self.server.building_blocks.remove(self.rect)
            Tree(self.server, self.x, self.y)
            self.kill()
            return

        for p in self.server.players:
            if not p.pending:
                screen = pygame.Rect(0, 0, 1000, 650)
                rect = pygame.Rect(0, 0, 1, 1)
                rect.size = self.rect.size
                rect.center = (p.character.get_x(self), p.character.get_y(self))
                if screen.colliderect(rect):
                    p.to_send.append({'action':'draw_obstacle',
                            'image':'sappling',
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'health':self.health,
                            'max_health':self.max_health})

    def getHurt(self, damage, attacker):
        if self.health > 0:
            self.health -= damage
            if self.health < 1:
                self.server.building_blocks.remove(self.innerrect)
                self.server.obs.remove(self)
                self.kill()

    def isBuilding(self):
        return False


    def explode(self):
        self.getHurt(120, None)


        

class Boulder(Obstacle):
    def __init__(self, server, x, y):
        Obstacle.__init__(self, server, x, y)
        self.image = 'boulder'
        self.max_health = self.health = r.randint(18, 32) * 10
        self.p = pygame.image.load('../assets/boulder.png')
        self.innerrect = self.p.get_rect(center=(x,y))
        server.building_blocks.append(self.innerrect)
        server.obs.append(self)


class Crate(Obstacle):
    def __init__(self, player, x, y):
        Obstacle.__init__(self, player.channel.server, x, y)
        self.image = 'crate'
        self.max_health = self.health = 800
        self.p = pygame.Surface((50, 50))
        self.innerrect = self.p.get_rect(center=(x,y))
        self.server.building_blocks.append(self.innerrect)
        self.server.obs.append(self)
        self.owner = player

    def explode(self):
        self.server.building_blocks.remove(self.innerrect)
        self.server.obs.remove(self)
        self.kill()

    def update(self):
        
        

        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'draw_obstacle',
                            'image':self.image,
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'health':self.health,
                            'max_health':self.max_health})


class Vine(Obstacle):
    def __init__(self, player, x, y, direction=None, speed=r.randint(25, 110)):
        Obstacle.__init__(self, player.channel.server, x, y)
        self.image = 'vine'
        self.max_health = self.health = speed * 2
        self.p = pygame.Surface((50, 50))
        self.innerrect = self.p.get_rect(center=(x,y))
        self.server.building_blocks.append(self.innerrect)
        self.server.obs.append(self)
        self.owner = player
        self.speed = speed
        self.count = self.speed

        
        dirs = [(0, 50), (50, 0), (-50, 0), (0, -50)]
        if direction:
            dirs.remove(direction)
        self.dir = r.choice(dirs)
        for v in self.server.obstacles:
            if v.__class__ == Vine and self.innerrect.colliderect(v.innerrect) and v != self:
                v.count = 1
                v.health = v.max_health
                v.dir = self.dir
                v.owner = self.owner
                self.explode()

    def explode(self):
        if self.innerrect in self.server.building_blocks:
            self.server.building_blocks.remove(self.innerrect)
        if self in self.server.obs:
            self.server.obs.remove(self)
        self.kill()

    def update(self):
        if self.count:
            self.count -= 1
            if not self.count:
                Vine(self.owner, self.x + self.dir[0], self.y + self.dir[1], direction=self.dir, speed=self.speed)
        

        for b in self.server.buildings:
            if not b.owner == self.owner.channel and self.innerrect.colliderect(b.innerrect):
                b.getHurt(0.02, self.owner.channel.username + '\'s invasive vine')
        
        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'draw_obstacle',
                            'image':self.image,
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'health':self.health,
                            'max_health':self.max_health})


class SpikyBush(pygame.sprite.Sprite):
    def __init__(self, player, x, y):
        pygame.sprite.Sprite.__init__(self, self.gp)
        self.server = player.channel.server
        self.x = x
        self.y = y
        self.dimensions = (50, 50)
        self.max_health = self.health = 260
        self.p = pygame.Surface(self.dimensions)
        self.rect = self.p.get_rect(center=(x,y))
        self.server.building_blocks.append(self.innerrect)
        self.owner = player

    @property
    def innerrect(self):
        return self.rect

    def explode(self):
        self.server.building_blocks.remove(self.innerrect)
        self.kill()

    def update(self):
        
        

        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'draw_obstacle',
                            'image':'spiky bush',
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'health':self.health,
                            'max_health':self.max_health})

    def getHurt(self, damage, attacker):
        if self.health > 0:
            self.health -= damage
            if self.health < 1:
                self.explode()

    def isBuilding(self):
        return False

class Gate(Obstacle):
    def __init__(self, player, x, y, rot):
        Obstacle.__init__(self, player.channel.server, x, y)
        self.image = 'gate'
        self.max_health = self.health = 1000
        if rot == False:
            self.p = pygame.Surface((100, 200))
        else:
            self.p = pygame.Surface((200, 100))
        self.innerrect = self.p.get_rect(center=(x,y))
        self.server.building_blocks.append(self.innerrect)
        self.server.obs.append(self)
        self.owner = player
        self.rotated = rot

    def explode(self):
        self.getHurt(900, None)

    def update(self):

        

        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'draw_obstacle',
                            'image':self.image,
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'health':self.health,
                            'max_health':self.max_health,
                            'rotated?':self.rotated})

class TNT(pygame.sprite.Sprite):
    def __init__(self, player, x, y):
        pygame.sprite.Sprite.__init__(self, self.gp)
        self.health = 130
        self.x = x
        self.y = y
        self.p = pygame.Surface((50, 50))
        self.innerrect = self.p.get_rect(center=(x,y))
        self.server = player.channel.server
        self.server.building_blocks.append(self.innerrect)
        self.server.obs.append(self)
        self.owner = player
        player.channel.achievement('You used TNT! (Watch out)')

    def update(self):
        self.health -= 1
        if self.health == 0:
            BAM(self)
            for p in self.server.players:
                screen = pygame.Rect(0, 0, 1000, 650)
                screen.center = p.character.rect.center
                if screen.colliderect(self.innerrect):
                    p.Send({'action':'sound', 'sound':'TNT'})
            self.server.building_blocks.remove(self.innerrect)
            self.server.obs.remove(self)
            
            ray = pygame.Rect(0, 0, 400, 400)
            ray.center = self.x, self.y
            for item in self.server.obstacles:
                if item != self and item.innerrect.colliderect(ray):
                    item.explode()
            for item in self.server.trees:
                if item != self and item.innerrect.colliderect(ray):
                    item.explode()
            for item in self.server.buildings:
                if item.innerrect.colliderect(ray):
                    item.getHurt(120, self.owner)
                elif item.rect.colliderect(ray):
                    item.getHurt(120, self.owner)
            for item in self.server.players:
                if item.character.rect.colliderect(ray):
                    item.character.getHurt(80, self.owner, t.getAngle(self.x, self.y, item.character.x, item.character.y), 120, '<Victim> was blown up by <Attacker>')
            for item in self.server.NPCs:
                if item.innerrect.colliderect(ray):
                    if item.__class__.__name__ == 'ArcheryTower':
                        item.explode(self.owner)
                    else:
                        item.explode(t.getAngle(self.x, self.y, item.x, item.y), 80)
            for item in self.server.event.NPCs:
                if item.rect.colliderect(ray):
                    item.explode(self.owner, self)

            for item in self.server.bushes:
                if item.rect.colliderect(ray):
                    item.explode()
                
            self.kill()

        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'draw_obstacle',
                            'image':'TNT',
                            'coords':(p.character.get_x(self), p.character.get_y(self))})
    def isBuilding(self):
        return False
    def getHurt(self, damage, attacker):
        pass
    def explode(self):
        ''' Ha! '''
        pass

class Block():
    def __init__(self, topleft, size):
        self.innerrect = pygame.Rect(topleft[0], topleft[1], size[0], size[1])
        self.owner = None
    def isBuilding(self):
        return False
    def getHurt(self, damage, attacker):
        pass
        
        
        




"""
rencode -- Web safe object pickling/unpickling. Python 3k version

The rencode module is a modified version of bencode from the
BitTorrent project.  For complex, heterogeneous data structures with
many small elements, r-encodings take up significantly less space than
b-encodings:

 >>> len(rencode.dumps({'a':0, 'b':[1,2], 'c':99}))
 13
 >>> len(bencode.bencode({'a':0, 'b':[1,2], 'c':99}))
 26

The rencode format is not standardized, and may change with different
rencode module versions, so you should check that you are using the
same rencode version throughout your project.
"""

__version__ = '1.0.0-ntk'
__all__ = ['dumps', 'loads', 'serializable']

from base64 import b64encode, b64decode

##TODO
# - Why the hell it encodes both tuples and lists to tuple?
#   Try loads(dumps([1,(2,3)]))
#   Grr
#
# - extend it! Support other basic types, f.e. Set()
#
##

# Original bencode module by Petru Paler, et al.
#
# Modifications by Joseph Parker:
# - Conversion to Python 3.x
#
# Modifications by Daniele Tricoli:
#
#  - Added support for instances
#    Only registered instances can be serialized. An instance to be serialized
#    must provide a '_pack` method.
#    E.g.
#
#    class X(object):
#
#        def __init__(self, x):
#            self.x = x
#
#        def _pack(self)
#           return (self.x,) # a tuple
#
#  - Lists are decoded again as list
#
# Modifications by Connelly Barnes:
#
#  - Added support for floats (sent as 32-bit or 64-bit in network
#    order), bools, None.
#  - Allowed dict keys to be of any serializable type.
#  - Lists/tuples are always decoded as tuples (thus, tuples can be
#    used as dict keys).
#  - Embedded extra information in the 'typecodes' to save some space.
#  - Added a restriction on integer length, so that malicious hosts
#    cannot pass us large integers which take a long time to decode.
#
# Licensed by Bram Cohen under the "MIT license":
#
#  "Copyright (C) 2001-2002 Bram Cohen
#
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation files
#  (the "Software"), to deal in the Software without restriction,
#  including without limitation the rights to use, copy, modify, merge,
#  publish, distribute, sublicense, and/or sell copies of the Software,
#  and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.
#
#  The Software is provided "AS IS", without warranty of any kind,
#  express or implied, including but not limited to the warranties of
#  merchantability,  fitness for a particular purpose and
#  noninfringement. In no event shall the  authors or copyright holders
#  be liable for any claim, damages or other liability, whether in an
#  action of contract, tort or otherwise, arising from, out of or in
#  connection with the Software or the use or other dealings in the
#  Software."
#
# (The rencode module is licensed under the above license as well).
#

import inspect
import struct

from threading import Lock
import types


class AlreadyRegistered(Exception): pass

class NotRegistered(Exception):

	def __init__(self, class_):
		self.class_ = class_

	def __str__(self):
		return 'Class %s is not registered' % self.class_

class NotSerializable(Exception): pass

def add_class_name(func):

	if inspect.ismethod(func):
		def decorate(*args, **kargs):
			result = func(*args, **kargs)
			result = (str(func.im_class.__name__),) + result
			return result

		return decorate

class _SerializableRegistry(object):

	def __init__(self):
		self._registry = {}

	def __contains__(self, item):
		return item in self._registry

	def __getitem__(self, key):
		return self._registry[key]

	def register(self, cls):
		'''   '''
		if inspect.isclass(cls):

			if cls.__name__ in self._registry:
				msg = 'Class %s is already registered' % cls.__name__
				raise AlreadyRegistered(msg)

			try:
				if inspect.ismethod(cls._pack):
					cls._pack = add_class_name(cls._pack)
					self._registry[cls.__name__] = cls
			except AttributeError as err:
				raise NotSerializable(err)

	def unregister(self, cls):
		'''   '''
		if inspect.isclass(cls):

			if cls in self._registry:
				del self._registry[cls.__name__]
			else: 
				raise NotRegistered(cls.__name__)

serializable = _SerializableRegistry()

# Number of bits for serialized floats, either 32 or 64.
FLOAT_BITS = 32

# Maximum length of integer when written as base 10 string.
MAX_INT_LENGTH = 64

# The bencode 'typecodes' such as i, d, etc have been extended and
# relocated on the base-256 character set.
# Can't be used chr(48) to chr(57) because they are manually set
CHR_INSTANCE = b'/'
CHR_TUPLE = b':'
CHR_LIST = b';'
CHR_DICT = b'<'
CHR_INT = b'='
CHR_INT1 = b'>'
CHR_INT2 = b'?'
CHR_INT4 = b'@'
CHR_INT8 = b'A'
CHR_FLOAT = b'B'
CHR_TRUE = b'C'
CHR_FALSE = b'D'
CHR_NONE = b'E' 
CHR_TERM = b'\x7f'

# Positive integers with value embedded in typecode.
INT_POS_FIXED_START = 0
INT_POS_FIXED_COUNT = 32

# Dictionaries with length embedded in typecode.
DICT_FIXED_START = 102
DICT_FIXED_COUNT = 25

# Negative integers with value embedded in typecode.
INT_NEG_FIXED_START = 70
INT_NEG_FIXED_COUNT = 32

# Strings with length embedded in typecode.
STR_FIXED_START = 128
STR_FIXED_COUNT = 64

# Lists with length embedded in typecode.
LIST_FIXED_START = STR_FIXED_START + STR_FIXED_COUNT
LIST_FIXED_COUNT = 32

# Tuples with length embedded in typecode.
TUPLE_FIXED_START = LIST_FIXED_START + LIST_FIXED_COUNT
TUPLE_FIXED_COUNT = 32

def decode_int(x, f):

	f += 1
	newf = x.index(CHR_TERM, f)
	if newf - f >= MAX_INT_LENGTH:
		raise ValueError('overflow')
	try:
		n = int(x[f:newf])
	except (OverflowError, ValueError):
		n = long(x[f:newf])
	if x[f] == '-':
		if x[f + 1] == '0':
			raise ValueError
	elif x[f] == '0' and newf != f+1:
		raise ValueError
	
	return (n, newf+1)

def decode_intb(x, f):
	f += 1
	return (struct.unpack('!b', x[f:f+1])[0], f+1)

def decode_inth(x, f):
	f += 1
	return (struct.unpack('!h', x[f:f+2])[0], f+2)

def decode_intl(x, f):
	f += 1
	return (struct.unpack('!l', x[f:f+4])[0], f+4)

def decode_intq(x, f):
	f += 1
	return (struct.unpack('!q', x[f:f+8])[0], f+8)

def decode_float(x, f):
	f += 1
	if FLOAT_BITS == 32:
		n = struct.unpack('!f', x[f:f+4])[0]
		return (n, f+4)
	elif FLOAT_BITS == 64:
		n = struct.unpack('!d', x[f:f+8])[0]
		return (n, f+8)
	else:
		raise ValueError

def decode_string(x, f):
	colon = x.index(b':', f)
	try:
		n = int(x[f:colon])
	except (OverflowError, ValueError):
		n = long(x[f:colon])
	if x[f] == '0' and colon != f+1:
		raise ValueError
	colon += 1
	return (x[colon:colon+n].decode(), colon+n)

def decode_list(x, f):
	r, f = [], f+1
	while x[f:f+1] != CHR_TERM:
		l = determine_flag_len(x, f)
		v, f = decode_func[x[f:f+l]](x, f)
		r.append(v)
	return (r, f + 1)

def decode_tuple(x, f):
	r, f = [], f+1
	while x[f:f+1] != CHR_TERM:
		l = determine_flag_len(x, f)
		v, f = decode_func[x[f:f+l]](x, f)
		r.append(v)
	return (tuple(r), f + 1)

def decode_dict(x, f):
	r, f = {}, f+1
	
	while x[f:f+1] != CHR_TERM:
		l = determine_flag_len(x, f)
		k, f = decode_func[x[f:f+l]](x, f)
		l = determine_flag_len(x, f)
		r[k], f = decode_func[x[f:f+l]](x, f)
	return (r, f + 1)

def decode_true(x, f):
  return (True, f+1)

def decode_false(x, f):
  return (False, f+1)

def decode_none(x, f):
  return (None, f+1)

def decode_instance(x, f):
	f += 1
	while x[f] != CHR_TERM:
		v, f = decode_func[x[f]](x, f)
	if v[0] in serializable:
		r = serializable[v[0]](*v[1:])
	else:
		raise NotRegistered(v[0])
	return (r, f+1)

decode_func = {}
decode_func[b'0'] = decode_string
decode_func[b'1'] = decode_string
decode_func[b'2'] = decode_string
decode_func[b'3'] = decode_string
decode_func[b'4'] = decode_string
decode_func[b'5'] = decode_string
decode_func[b'6'] = decode_string
decode_func[b'7'] = decode_string
decode_func[b'8'] = decode_string
decode_func[b'9'] = decode_string
decode_func[CHR_LIST ] = decode_list
decode_func[CHR_TUPLE] = decode_tuple
decode_func[CHR_DICT ] = decode_dict
decode_func[CHR_INT  ] = decode_int
decode_func[CHR_INT1 ] = decode_intb
decode_func[CHR_INT2 ] = decode_inth
decode_func[CHR_INT4 ] = decode_intl
decode_func[CHR_INT8 ] = decode_intq
decode_func[CHR_FLOAT] = decode_float
decode_func[CHR_TRUE ] = decode_true
decode_func[CHR_FALSE] = decode_false
decode_func[CHR_NONE ] = decode_none
decode_func[CHR_INSTANCE] = decode_instance

def make_fixed_length_string_decoders():
	def make_decoder(slen):
		def f_fixed_string(x, f):
			l = determine_flag_len(x, f)
			return (x[f+l:f+l+slen].decode(), f+l+slen)
		return f_fixed_string
	for i in range(STR_FIXED_COUNT):
		#print(chr(STR_FIXED_START+i).encode())
		decode_func[chr(STR_FIXED_START+i).encode()] = make_decoder(i)

make_fixed_length_string_decoders()

## switch for type identifier, is it 1 or 2 bytes?
def determine_flag_len(x, f): #data, index
	l = 1 #length
	if x[f:f+1] in (b'\xc2', b'\xc3'): #have to slice to preserve b'bytes'
		l = 2
	return l


def make_fixed_length_list_decoders():
	def make_decoder(slen):
		def f_fixed_list(x, f):
			l = determine_flag_len(x, f)
			r, f = [], f+l
			
			for i in range(slen):
				l = determine_flag_len(x, f)
				v, f = decode_func[x[f:f+l]](x, f)
				r.append(v)
			return (r, f)
		return f_fixed_list
	for i in range(LIST_FIXED_COUNT):
		decode_func[chr(LIST_FIXED_START+i).encode()] = make_decoder(i)

make_fixed_length_list_decoders()

def make_fixed_length_tuple_decoders():
	def make_decoder(slen):
		def f_fixed_tuple(x, f):
			l = determine_flag_len(x, f)
			r, f = [], f+l
			for i in range(slen):
				l = determine_flag_len(x, f)
				v, f = decode_func[x[f:f+l]](x, f)
				r.append(v)
			return (tuple(r), f)
		return f_fixed_tuple
	for i in range(TUPLE_FIXED_COUNT):
		decode_func[chr(TUPLE_FIXED_START+i).encode()] = make_decoder(i)

make_fixed_length_tuple_decoders()

def make_fixed_length_int_decoders():
	def make_decoder(j):
		def f(x, f):
			return (j, f+1)
		return f
	for i in range(INT_POS_FIXED_COUNT):
		decode_func[chr(INT_POS_FIXED_START+i).encode()] = make_decoder(i)
	for i in range(INT_NEG_FIXED_COUNT):
		decode_func[chr(INT_NEG_FIXED_START+i).encode()] = make_decoder(-1-i)

make_fixed_length_int_decoders()

def make_fixed_length_dict_decoders():
	def make_decoder(slen):
		def f(x, f):
			l = determine_flag_len(x, f)
			r, f = {}, f+l
			
			for j in range(slen):
				
				l = determine_flag_len(x, f)
				k, f = decode_func[x[f:f+l]](x, f)
				
				l = determine_flag_len(x, f)
				r[k], f = decode_func[x[f:f+l]](x, f)
				
				
			return (r, f)
			
		return f
	for i in range(DICT_FIXED_COUNT):
		decode_func[chr(DICT_FIXED_START+i).encode()] = make_decoder(i)

make_fixed_length_dict_decoders()

def loads(x):
	flagl = determine_flag_len(x, 0)
	try:
		r, l = decode_func[x[0:flagl]](x, 0)
	except (IndexError, KeyError):
		raise 
	if l != len(x):
		print(r, l, len(x))
		print("ValueError")
		#raise ValueError
	return r

def encode_int(x, r):
	if 0 <= x < INT_POS_FIXED_COUNT:
		r.append(chr(INT_POS_FIXED_START+x).encode())
		
	elif -INT_NEG_FIXED_COUNT <= x < 0:
		r.append(chr(INT_NEG_FIXED_START-1-x).encode())
	elif -128 <= x < 128:
		r += [CHR_INT1, struct.pack('!b', x)]
		
	elif -32768 <= x < 32768:
		r += [CHR_INT2, struct.pack('!h', x)]
		
	elif -2147483648 <= x < 2147483648:
		r += [CHR_INT4, struct.pack('!l', x)]
	elif -9223372036854775808 <= x < 9223372036854775808:
		
		r += [CHR_INT8, struct.pack('!q', x)]
	else:
		s = str(x).encode()
		if len(s) >= MAX_INT_LENGTH:
			raise ValueError('overflow')
		r += [CHR_INT, s, CHR_TERM]

def encode_float(x, r):

	if FLOAT_BITS == 32:

		r += [CHR_FLOAT, struct.pack('!f', x)]
	elif FLOAT_BITS == 64:

		r += [CHR_FLOAT, struct.pack('!d', x)]
	else:
		raise ValueError

def encode_bool(x, r):

	r += [{False: CHR_FALSE, True: CHR_TRUE}[bool(x)] ]

def encode_none(x, r):
	r.append(CHR_NONE)

def encode_string(x, r):
	x = x.encode()
	if len(x) < STR_FIXED_COUNT:
		r += [chr(STR_FIXED_START + len(x)).encode(), x]
	else:
		r += [str(len(x)).encode(), b':', x]

def encode_list(x, r):
	if len(x) < LIST_FIXED_COUNT:
		r += [chr(LIST_FIXED_START + len(x)).encode()]
		for i in x:
			encode_func.get(type(i), encode_instance)(i, r)
	else:
		r.append(CHR_LIST)
		for i in x:
			encode_func.get(type(i), encode_instance)(i, r)
		r.append(CHR_TERM)


def encode_tuple(x, r):
	if len(x) < TUPLE_FIXED_COUNT:
		r += [chr(TUPLE_FIXED_START + len(x)).encode() ]
		for i in x:
			encode_func.get(type(i), encode_instance)(i, r)
	else:
		r.append(CHR_TUPLE)
		for i in x:
			encode_func.get(type(i), encode_instance)(i, r)
		r.append(CHR_TERM)

def encode_dict(x,r):
	if len(x) < DICT_FIXED_COUNT:
		r.append(chr(DICT_FIXED_START + len(x)).encode() )
		for k, v in x.items():
			encode_func[type(k)](k, r)
			encode_func[type(v)](v, r)
	else:
		r.append(CHR_DICT) 
		for k, v in x.items():
			encode_func[type(k)](k, r)
			encode_func[type(v)](v, r)
		r.append(CHR_TERM) 
	

encode_func = {}
encode_func[int] = encode_int
encode_func[int] = encode_int
encode_func[float] = encode_float
encode_func[str] = encode_string
encode_func[list] = encode_list
encode_func[tuple] = encode_tuple
encode_func[dict] = encode_dict
encode_func[type(None)] = encode_none
encode_func[type(True)] = encode_bool


def encode_instance(x, r):
	if hasattr(x, '_pack'):
		if x.__class__.__name__ in serializable:
			# Calling the class of instance `x' passing it to the
			# unbound method
			result = serializable[x.__class__.__name__]._pack(x)
			r.append(CHR_INSTANCE.encode())
			encode_func[type(result)](result, r)
			r.append(CHR_TERM.encode())
		else:
			raise NotRegistered(x.__class__.__name__)

lock = Lock()

def dumps(x):
	lock.acquire()
	r = []
	encode_func.get(type(x), encode_instance)(x, r)
	lock.release()
	for entry in r:
		if type(entry) != bytes:
			print("!!!!!!!!!", entry)
	return b''.join(r)

def test():
	f1 = struct.unpack('!f', struct.pack('!f', 25.5))[0]
	f2 = struct.unpack('!f', struct.pack('!f', 29.3))[0]
	f3 = struct.unpack('!f', struct.pack('!f', -0.6))[0]
	L = (({'a':15, 'bb':f1, 'ccc':f2, '':(f3,(),False,True,'')},('a',10**20),tuple(range(-100000,100000)),'b'*31,'b'*62,'b'*64,2**30,2**33,2**62,2**64,2**30,2**33,2**62,2**64,False,False, True, -1, 2, 0),)
	G = loads(dumps(L))
	
	assert G == L
	d = dict(zip(range(-100000,100000),range(-100000,100000)))
	d.update({'a':20, 20:40, 40:41, f1:f2, f2:f3, f3:False, False:True, True:False})
	L = (d, {}, {5:6}, {7:7,True:8}, {9:10, 22:39, 49:50, 44: ''})
	assert loads(dumps(L)) == L
	L = ('', 'a'*10, 'a'*100, 'a'*1000, 'a'*10000, 'a'*100000, 'a'*1000000, 'a'*10000000)
	assert loads(dumps(L)) == L
	L = tuple([dict(zip(range(n),range(n))) for n in range(100)]) + ('b',)
	assert loads(dumps(L)) == L
	L = tuple([dict(zip(range(n),range(-n,0))) for n in range(100)]) + ('b',)
	assert loads(dumps(L)) == L
	L = tuple([tuple(range(n)) for n in range(100)]) + ('b',)
	assert loads(dumps(L)) == L
	L = tuple(['a'*n for n in range(100)]) + ('b',)
	assert loads(dumps(L)) == L
	L = tuple(['a'*n for n in range(100)]) + (None,True,None)
	assert loads(dumps(L)) == L
	L = list(['a'*n for n in range(100)]) + [None,True,None]
	assert loads(dumps(L)) == L
	assert loads(dumps(None)) == None
	assert loads(dumps({None:None})) == {None:None}

	class A(object):
		def __init__(self, a, b, c):
			self.a = a
			self.b = b
			self.c = c

		def _pack(self):
			return (self.a, self.b, self.c)

	serializable.register(A)

	instance = [A(1,2,3), 1, A(1,3,4), 'sss']
	print(loads(dumps(instance)))

if __name__ == '__main__':
  test()


import pygame
import random as r
import toolbox as t
from balloon import Arrow
from animations import *

class Farmer(pygame.sprite.Sprite):
    def __init__(self, building):
        pygame.sprite.Sprite.__init__(self, self.gp)
        self.building = building
        self.x, self.y = building.rect.x, building.rect.y
        self.setout()
        self.speed = 2
        self.obs = []
        self.food = 0

        self.rect = None

    @property
    def innerrect(self):
        return self.rect

    def setout(self):
        self.status = 'Going to Farms'
        sm_dist = 8000
        for farm in self.building.server.resources:
            if farm.__class__.__name__ == 'Farm':
                dx, dy = r.randint(farm.x + 90, farm.x + 310), farm.y + 50
                dist = t.getDist(self.x, self.y, dx, dy)
                if dist < sm_dist:
                    sm_dist = dist
                    x, y = dx, dy
                    self.farm = farm
        self.dest_x = x
        self.dest_y = y
        self.angle = t.getAngle(self.x, self.y, x, y)

    def start_gather(self):
        self.status = 'Gathering Food'
        sm_dist = 1000
        for food in self.farm.foods:
            if food.image == 'good':
                dist = t.getDist(self.x, self.y, food.x, food.y)
                if dist < sm_dist:
                    sm_dist = dist
                    self.the_food = food
                    self.dest_x = food.x
                    self.dest_y = food.y

    def mine(self):
        try:
            self.the_food.mine(self)
            if self.the_food.image == 'used':
                self.start_gather()
        except:
            self.start_gather()

    def at_dest(self):
        return t.getDist(self.x, self.y, self.dest_x, self.dest_y) < 10

    def go_back(self):
        self.status = 'Returning to Guild'
        self.dest_x, self.dest_y = self.building.rect.center

    def update(self):

        
        self.obs = self.building.server.obs
        
        self.angle = t.getAngle(self.x, self.y, self.dest_x, self.dest_y)
        x, y = t.getDir(self.angle, self.speed)
        
        lx, ly = self.x, self.y
        if self.status != 'Gathering Food':
            self.x += x
            self.rect = pygame.Rect(self.x, self.y, 50, 50)
            for ob in self.obs:
                if self.rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.building.owner):
                    self.x -= x
            
            self.y += y
            self.rect = pygame.Rect(self.x, self.y, 50, 50)
            for ob in self.obs:
                if self.rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.building.owner):
                    self.y -= y
            if round(lx) == round(self.x) and round(ly) == round(self.y) and self.status != 'Gathering Food':
                blocked = True
                while blocked:
                    blocked = False
                    
                    self.x += x
                    self.rect = pygame.Rect(self.x, self.y, 50, 50)
                    for ob in self.obs:
                        if self.rect.colliderect(ob.innerrect):
                            blocked = True
            
                    self.y += y
                    self.rect = pygame.Rect(self.x, self.y, 50, 50)
                    for ob in self.obs:
                        if self.rect.colliderect(ob.innerrect):
                            blocked = True
                

        if self.status == 'Gathering Food':
            if self.food < 20:
                self.mine()
            else:
                self.go_back()
            

        if self.at_dest():
            if self.status == 'Going to Farms':
                self.start_gather()
            elif self.status == 'Returning to Guild':
                self.food = 0
                self.building.owner.character.food += 20
                self.setout()
        
        for p in self.building.server.players:
            if not p.pending:
                screen = pygame.Rect(0, 0, 1000, 650)
                rect = pygame.Rect(0, 0, 1, 1)
                rect.size = self.rect.size
                rect.topleft = (p.character.get_x(self.rect), p.character.get_y(self.rect))
                if screen.colliderect(rect):
                    if self.status == 'Gathering Food':
                        p.to_send.append({'action':'draw_NPC',
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'image':'farmer',
                            'status':'Gathering Food : ' + str(self.food) + '/20',
                            'color':self.building.owner.color,
                            'angle':self.angle})
                    else:
                        p.to_send.append({'action':'draw_NPC',
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'image':'farmer',
                            'status':self.status,
                            'color':self.building.owner.color,
                            'angle':self.angle})
    def explode(self, angle, knockback):
        x, y = t.getDir(angle, knockback)
        self.x += x
        self.rect = pygame.Rect(self.x, self.y, 50, 50)
        for ob in self.obs:
            if self.rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.building.owner):
                self.x -= x
            
        self.y += y
        self.rect = pygame.Rect(self.x, self.y, 50, 50)
        for ob in self.obs:
            if self.rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.building.owner):
                self.y -= y


class Miner(pygame.sprite.Sprite):
    def __init__(self, building):
        pygame.sprite.Sprite.__init__(self, self.gp)
        self.building = building
        self.x, self.y = building.rect.x, building.rect.y
        self.setout()
        self.speed = 2
        self.gold = 0
        self.obs = []

    @property
    def innerrect(self):
        return self.rect        

    def setout(self):
        self.status = 'Heading to Mines'
        sm_dist = 8000
        for farm in self.building.server.resources:
            if farm.__class__.__name__ == 'Mine':
                dx, dy = farm.x+100, r.randint(farm.y + 50, farm.y + 400)
                dist = t.getDist(self.x, self.y, dx, dy)
                if dist < sm_dist:
                    sm_dist = dist
                    x, y = dx, dy
                    self.farm = farm
        self.dest_x = x
        self.dest_y = y
        self.angle = t.getAngle(self.x, self.y, x, y)

    def start_gather(self):
        self.status = 'Mining'
        sm_dist = 1000
        for food in self.farm.golds:
            if food.image == 'good':
                dist = t.getDist(self.x, self.y, food.x, food.y)
                if dist < sm_dist:
                    sm_dist = dist
                    self.the_food = food
                    self.dest_x = food.x
                    self.dest_y = food.y

    def mine(self):
        try:
            self.the_food.collect(self)
            if self.the_food.image == 'used':
                self.start_gather()
        except:
            self.start_gather()

    def at_dest(self):
        return t.getDist(self.x, self.y, self.dest_x, self.dest_y) < 10

    def go_back(self):
        self.status = 'Returning to Guild'
        self.dest_x, self.dest_y = self.building.rect.center

    def update(self):
        self.obs = self.building.server.obs
        self.angle = t.getAngle(self.x, self.y, self.dest_x, self.dest_y)
        x, y = t.getDir(self.angle, self.speed)
        
        lx, ly = self.x, self.y
        if self.status != 'Mining':
            self.x += x
            self.rect = pygame.Rect(self.x, self.y, 50, 50)
            for ob in self.obs:
                if self.rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.building.owner):
                    self.x -= x
            
            self.y += y
            self.rect = pygame.Rect(self.x, self.y, 50, 50)
            for ob in self.obs:
                if self.rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.building.owner):
                    self.y -= y
            if round(lx) == round(self.x) and round(ly) == round(self.y) and self.status != 'Mining':
                blocked = True
                while blocked:
                    blocked = False
                    
                    self.x += x
                    self.rect = pygame.Rect(self.x, self.y, 50, 50)
                    for ob in self.obs:
                        if self.rect.colliderect(ob.innerrect):
                            blocked = True
            
                    self.y += y
                    self.rect = pygame.Rect(self.x, self.y, 50, 50)
                    for ob in self.obs:
                        if self.rect.colliderect(ob.innerrect):
                            blocked = True
                

        if self.status == 'Mining':
            if self.gold < 20:
                self.mine()
            else:
                self.go_back()
            

        if self.at_dest():
            if self.status == 'Heading to Mines':
                self.start_gather()
            elif self.status == 'Returning to Guild':
                self.gold = 0
                self.building.owner.character.gold += 20
                self.setout()
        
        for p in self.building.server.players:
            if not p.pending:
                screen = pygame.Rect(0, 0, 1000, 650)
                rect = pygame.Rect(0, 0, 1, 1)
                rect.size = self.rect.size
                rect.topleft = (p.character.get_x(self.rect), p.character.get_y(self.rect))
                if screen.colliderect(rect):
                    if self.status == 'Mining':
                        p.to_send.append({'action':'draw_NPC',
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'image':'miner',
                            'status':'Mining : ' + str(self.gold) + '/20',
                            'color':self.building.owner.color,
                            'angle':self.angle})
                    else:
                        p.to_send.append({'action':'draw_NPC',
                            'coords':(p.character.get_x(self), p.character.get_y(self)),
                            'image':'miner',
                            'status':self.status,
                            'color':self.building.owner.color,
                            'angle':self.angle})
                        
    def explode(self, angle, knockback):
        x, y = t.getDir(angle, knockback)
        self.x += x
        self.rect = pygame.Rect(self.x, self.y, 50, 50)
        for ob in self.obs:
            if self.rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.building.owner):
                self.x -= x
            
        self.y += y
        self.rect = pygame.Rect(self.x, self.y, 50, 50)
        for ob in self.obs:
            if self.rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.building.owner):
                self.y -= y


class ArcheryTower(pygame.sprite.Sprite):
    dimensions = (360, 360)
    def __init__(self, server, x, y, channel, rect):
        pygame.sprite.Sprite.__init__(self, self.gp)
        self.building = self
        
        self.server = server
        self.x, self.y = x, y - 100
        self.owner = channel.character
        self.health = 300
        self.dimensions = (360, 360)
        
        self.innerrect = rect
        self.server.building_blocks.append(self.innerrect)
        self.server.obs.append(self)

        self.balloon_speed = 45
        self.attack = 24
        self.knockback = 32

        self.shoot_cooldown = 60
        self.state = 'alive'

        self.attacking = False

    def update(self):
        self.lookx = self.owner.x
        self.looky = self.owner.y
        self.attacking = False

        for npc in self.server.event.NPCs:
            if t.getDist(self.x, self.y, npc.x, npc.y) < 800:
                    self.lookx = npc.x
                    self.looky = npc.y
                    self.attacking = True

                    
        for player in self.server.players:
            if not player.pending:
                if t.getDist(self.x, self.y, player.character.x, player.character.y) < 800 and player != self.owner.channel and player.character.dead == False:
                    self.lookx = player.character.x
                    self.looky = player.character.y
                    self.attacking = True

        
                
        if self.attacking == False:
            close_dist = 100000
            
            for robot in self.server.NPCs:
                if robot.__class__.__name__ == 'Robot' and robot.factory.owner != self.owner.channel and t.getDist(self.x, self.y, robot.x, robot.y) < close_dist:
                    self.lookx = robot.x
                    self.looky = robot.y
                    close_dist = t.getDist(self.x, self.y, robot.x, robot.y)
                    self.attacking = True

        self.angle = t.getAngle(self.x, self.y, self.lookx, self.looky)
        if self.shoot_cooldown > 0:
            self.shoot_cooldown -= 1

        if self.shoot_cooldown < 1 and self.attacking and self.state == 'alive':
            Arrow(self)
            self.shoot_cooldown = 60


        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'archery_tower',
                    'coords':(p.character.get_x(self), p.character.get_y(self)),
                    'angle':self.angle,
                    'color':self.owner.channel.color,
                    'health':self.health,
                    'state':self.state})


    def isBuilding(self):
        return False
    
    def getHurt(self, damage, attacker):
        if attacker != self.owner:
            self.health -= damage
            if self.health < 1:
                self.state = 'broken'
                if attacker.__class__.__name__ == 'Character':
                    attacker.destroyed += 1
                    self.owner.channel.message = attacker.channel.username + ' has broken one of your Archery Towers.'
                else:
                    self.owner.channel.message = attacker + ' has broken one of your Archery Towers.'
                    
                self.owner.channel.message_count = 150
                self.owner.channel.message_color = (255,0,0)
                self.server.obs.remove(self)

    def explode(self, player):
        if self.state == 'broken':
            self.server.building_blocks.remove(self.innerrect)
            self.kill()
        else:
            self.getHurt(80, player)
            if self.state == 'broken':
                self.server.building_blocks.remove(self.innerrect)
                self.kill()





class Robot(pygame.sprite.Sprite):
    def __init__(self, factory):
        pygame.sprite.Sprite.__init__(self, self.gp)

        self.server = factory.server
        self.factory = factory
        self.player = factory.owner.character
        self.angle = r.randint(0, 359)

        self.x = self.factory.x
        self.y = self.factory.y + 140
        
        self.fav_dir = r.choice((5, -5))
        self.angry_test = 20
        self.regular_test = 2
        self.angry = False

        self.speed = 4
        self.health = 100
        self.hurt = 0

        self.dead = 0

        self.obs = []

        self.state = 'Peaceful'
        self.surf = pygame.Surface((50, 50))
        self.rect = self.surf.get_rect(center=(self.x, self.y))

        self.safety_zone = pygame.Surface((3400, 2190)).get_rect(center=(self.factory.x, self.factory.y))

        self.range = pygame.Surface((3000, 2000)).get_rect(center=(self.factory.x, self.factory.y))

    @property
    def innerrect(self):
        return self.rect

    @property
    def building(self):
        return self.factory


        

    def update(self):
        rect = self.surf.get_rect(center=(self.x, self.y))
        for ob in self.obs:
            if rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.factory.owner):
                self.getHurt(100, 0, 0)

        
        self.obs = self.server.obs
        for robot in self.server.NPCs:
            if robot.__class__.__name__ == 'Robot' and robot.factory.owner != self.factory.owner:
                self.obs.append(robot)

        if self.hurt:
            self.hurt -= 1

        if self.dead:
            self.dead -= 1

        if not self.dead:
            self.AI()

            for p in self.server.players:
                if not p.pending:
                    screen = pygame.Rect(0, 0, 1000, 650)
                    rect = pygame.Rect(0, 0, 1, 1)
                    rect.size = self.rect.size
                    rect.topleft = (p.character.get_x(self.rect), p.character.get_y(self.rect))
                    if screen.colliderect(rect):
                        p.to_send.append({'action':'draw_robot',
                                'coords':(p.character.get_x(self), p.character.get_y(self)),
                                'name':'Robot',
                                'health':self.health,
                                'image':('hurt' if self.hurt else 'regular'),
                                'color':self.factory.owner.color,
                                'angle':self.angle})


            for p in self.server.players:
                if not p.pending and self.rect.colliderect(p.character.rect) and p != self.factory.owner and p.character.dead == False:
                    self.getHurt(0, self.angle + 180, 50)
                    p.character.getHurt(10, self.factory.owner.username + "'s Robots", self.angle, 2)


    def AI(self):

        if not self.rect.colliderect(self.safety_zone):
            self.state = 'Returning'
            self.angle = t.getAngle(self.x, self.y, self.factory.x, self.factory.y + 140)
        
        if self.state == 'Returning':
            if round(self.x) == round(self.factory.x) and round(self.y) == round(self.factory.y):
                self.state = 'Peaceful'

        if self.state == 'Attacking':
            speed = self.speed
        else:
            speed = 2


        origx = self.x

        
        x, y = t.getDir(self.angle, total=speed)

        self.x += x
        rect = self.surf.get_rect(center=(self.x, self.y))
        for ob in self.obs:
            if rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.factory.owner):
                self.x -= x

        origy = self.y
            
        self.y += y
        rect = self.surf.get_rect(center=(self.x, self.y))
        for ob in self.obs:
            if rect.colliderect(ob.innerrect) and (ob.__class__.__name__ != 'Gate' or ob.owner.channel != self.factory.owner):
                self.y -= y

        for item in self.server.bushes:
            if self.rect.colliderect(item.innerrect):
                if item.__class__.__name__ == 'SpikyBush':
                    self.getHurt(1, self.angle + 180, 1.5)
                else:
                    self.getHurt(0, self.angle + 180, 1.5)
            
        self.rect = self.surf.get_rect(center=(self.x, self.y))

        suspects = []
        for p in self.server.players:
            if not p.pending:
                if p.character.rect.colliderect(self.range) and not p.character.dead and not p == self.factory.owner:
                    suspects.append(p.character)
        if len(suspects) > 1:
            closest = suspects[0]
            for suspect in suspects[1:]:
                if t.getDist(self.x, self.y, suspect.x, suspect.y) < t.getDist(self.x, self.y, closest.x, closest.y):
                    closest = suspect
                    suspect = closest
        elif len(suspects) == 1:
            suspect = suspects[0]

        else:
            suspect = None

        if suspect != None:
            self.state = 'Attacking'
            self.attacking = suspect
            self.angle = t.getAngle(self.x, self.y, self.attacking.x, self.attacking.y) + r.randint(-2,2)


    def init(self):

        self.x = self.factory.x
        self.y = self.factory.y + 140

        self.state = 'Peaceful'
        
        self.health = 100

    def getHurt(self, damage, angle, knockback):

        self.health -= damage
        if self.health <= 0:
            Explosion(self)
            self.dead = 200
            for p in self.server.players:
                if not p.pending:
                    screen = pygame.Rect(0, 0, 1000, 650)
                    screen.center = p.character.rect.center
                    if screen.colliderect(self.rect):
                        p.to_send.append({'action':'sound', 'sound':'die'})

            self.init()

        if damage != 0:
            self.hurt = 8
        
        x, y = t.getDir(angle, knockback)

        self.x += x
        rect = self.surf.get_rect(center=(self.x, self.y))
        for ob in self.obs:
            if rect.colliderect(ob.innerrect):
                self.x -= x


        
        self.y += y
        rect = self.surf.get_rect(center=(self.x, self.y))
        for ob in self.obs:
            if rect.colliderect(ob.innerrect):
                self.y -= y

    def explode(self, angle, knockback):
        self.getHurt(80, angle, knockback)

















    

                
        



from building import *

def kick(self, playername):
    for p in self.server.players:
        if p.username == playername:
            p.Send({'action':'kicked'})
    
    print('Kicked ' + playername)
    del self.server.players[playername]


def makeWallsFall(self):
    self.server.upwall.count = 10
    self.server.leftwall.count = 10


def kill(self, playername):
    for p in self.server.players:
        if p.username == playername:
            break
    p = p.character
    p.getHurt(300, self.character, 0, 0, msg="<Attacker> killed <Victim> using cheats")


def win(self):
    self.server.terminate(self)


def message(self, theMessage, color=(0,0,0)):
    for p in self.server.players:
        p.message = message
        p.message_color = color
        p.message_count = 160



import pygame
import pygame as p
from pygame.locals import *

import random
import time
import threading
import re
import sys
import os
from pymsgbox import alert, confirm, prompt, password

try:
    from PodSixNet.Connection import ConnectionListener, connection
    import toolbox as t
except:
    from src.PodSixNet.Connection import ConnectionListener, connection
    from src import toolbox as t

toolbox = t

answer = ''

def hack():
    command = prompt('Enter the command:', title='VillageWars Console')
    if command is not None:
        connection.Send({'action':'hack', 'command':command})

def find_cursor(cursor_param):
    regex = re.compile(r'constant: SYSTEM_CURSOR_(.*)\)>')
    res = regex.search(repr(cursor_param))
    return res.group(1)


class MyGameClient(ConnectionListener):
    def __init__(self, host, port, screen, clock, username, version, color, skin):

        """
        Client constructor function. This is the code that runs once
        when a new client is made.
        """
        ConnectionListener.__init__(self)

        self.color = color
        # Start the game
        pygame.init()
        pygame.mixer.pre_init(buffer=64)
        game_width = 1000
        game_height = 650
        self.started = False
        self.screen = screen
        self.clock = clock
        self.skin = skin

        self.fps_expected = 30

        self.toDraw = []
        self.toDrawNext = []

        self.disconnected = 0

        self.back_pic = p.image.load('../assets/BG_SciFi.png')
        self.ocean_pics = [p.transform.scale(p.image.load('../assets/ocean/' + i), (1000, 1000)) for i in os.listdir('../assets/ocean/')]
        self.ocean_frame = 0
        self.players = [(p.transform.scale(p.image.load('../assets/Skins/%s.png' % (i)), (50, 50)), p.transform.scale(p.image.load('../assets/Skins/%sh.png' % (i)), (50, 50))) for i in range(len(os.listdir('../assets/Skins'))//2)]
        self.archer = p.transform.scale(p.image.load('../assets/Enemy_04.png'), (50, 50))

        self.bolt = p.image.load('../assets/barbarians/bolt.png')

        self.startgame_icon = p.image.load('../assets/Startgame_btn.png')
        self.startgame_icon_over = p.image.load('../assets/Startgame_btn_over.png')
        self.startgame_icon_locked = p.image.load('../assets/Startgame_btn_locked.png')

        self.balloon = p.image.load('../assets/balloon.png')
        self.op_balloon = p.image.load('../assets/balloon2.png')
        self.speedy_plus_op = p.image.load('../assets/Drop.png')
        self.speedy_balloon = p.image.load('../assets/DropSmall.png')
        self.robot = p.transform.scale(p.image.load('../assets/Enemy_05.png'), (50, 50))
        self.robot_hurt = p.transform.scale(p.image.load('../assets/Enemy_05_hurt.png'), (50, 50))

        self.animations = {
            'Splash':(p.image.load('../assets/splash1.png'), p.image.load('../assets/splash2.png'), p.image.load('../assets/splash3.png')),
            'Explosion':(p.image.load('../assets/explode1.png'), p.image.load('../assets/explode2.png'), p.image.load('../assets/explode3.png')),
            'BAM':(p.transform.scale(p.image.load('../assets/LargeExplosion1.png'), (360, 360)), p.transform.scale(p.image.load('../assets/LargeExplosion1.png'), (360, 360)), p.transform.scale(p.image.load('../assets/LargeExplosion2.png'), (360, 360)), p.transform.scale(p.image.load('../assets/LargeExplosion2.png'), (360, 360)), p.transform.scale(p.image.load('../assets/LargeExplosion3.png'), (360, 360)), p.transform.scale(p.image.load('../assets/LargeExplosion3.png'), (360, 360)))
            }

        self.paused_image = p.image.load('../assets/paused.png')
        
        self.red = p.image.load('../assets/buildings/red.png')
        self.blue = p.image.load('../assets/buildings/blue.png')
        self.green = p.image.load('../assets/buildings/green.png')
        self.yellow = p.image.load('../assets/buildings/yellow.png')
        self.preview = p.image.load('../assets/buildings/preview.png')

        self.house = p.image.load('../assets/buildings/House.png')
        self.house_burnt = p.image.load('../assets/buildings/House_burnt.png')

        self.building_person = p.transform.scale(p.image.load('../assets/Skins/4.png'), (50, 50))
  
        self.miner = p.transform.scale(p.image.load('../assets/Skins/2.png'), (50, 50))
        self.farmer = p.transform.scale(p.image.load('../assets/Skins/4.png'), (50, 50))

        self.barbarians = {
            'leader':[p.image.load('../assets/barbarians/leader.png'), p.image.load('../assets/barbarians/leaderh.png'), p.image.load('../assets/barbarians/leaderm.png')],
            'archer':[p.image.load('../assets/barbarians/archer.png'), p.image.load('../assets/barbarians/archerh.png')],
            'swordsman':[p.image.load('../assets/barbarians/swordsman.png'), p.image.load('../assets/barbarians/swordsmanh.png')]
            }
        self.shield = p.image.load('../assets/barbarians/shield.png')
        self.barbarian_banner = p.transform.scale(p.image.load('../assets/barbarians/banner.png'), (45, 90))

        for i, b in enumerate(self.barbarians['leader']):
            self.barbarians['leader'][i] = p.transform.scale(b, (50, 50))
        for i, b in enumerate(self.barbarians['archer']):
            self.barbarians['archer'][i] = p.transform.scale(b, (50, 50))
        for i, b in enumerate(self.barbarians['swordsman']):
            self.barbarians['swordsman'][i] = p.transform.scale(b, (50, 50))
        
        self.startgame_rect = self.startgame_icon.get_rect(topleft=(445, 400))

        self.Setting = p.image.load('../assets/Setting_walls.png')

        self.food_image = p.image.load('../assets/food.png')
        self.food_used = p.image.load('../assets/food_used.png')
        self.food_mine = p.image.load('../assets/food_mine.png')

        self.plate = p.transform.scale(p.image.load('../assets/meals/plate.png'), (100, 100))
        self.meals = [
            p.transform.scale(p.image.load('../assets/meals/meal1.png'), (100, 100)),
            p.transform.scale(p.image.load('../assets/meals/meal2.png'), (100, 100)),
            p.transform.scale(p.image.load('../assets/meals/meal3.png'), (100, 100)),
            p.transform.scale(p.image.load('../assets/meals/meal4.png'), (100, 100)),
                     ]
        self.meals = [p.transform.scale(p.image.load('../assets/meals/meal%s.png' % (num - 1)), (100, 100)) for num in range(len(os.listdir('../assets/meals'))) if num not in (0, 1)]

        self.gold_image = p.transform.scale(p.image.load('../assets/gold.png'), (45, 45))
        self.gold_mine = p.transform.scale(p.image.load('../assets/gold_mine.png'), (45, 45))

        self.mine = p.image.load('../assets/mine.png')

        self.x = p.image.load('../assets/x.png')
        self.x_hov = p.image.load('../assets/x_hov.png')

        self.tree = p.transform.scale(p.image.load('../assets/trees/tree1.png'), (200, 280))
        self.sappling = p.image.load('../assets/trees/sappling.png')

        self.large_font = p.font.SysFont('default', 80)
        self.medium_font = p.font.SysFont('default', 40)
        self.small_font = p.font.SysFont('default', 20)

        self.crate = p.transform.scale(p.image.load('../assets/crate.png'), (50, 50))
        
        self.tnt = p.transform.scale(p.image.load('../assets/TNT.png'), (50, 50))
        self.gate = p.image.load('../assets/gate.png')
        self.spiky_bush = p.image.load('../assets/spiky_bush.png')

        self.ready_text = self.small_font.render('READY', True, (0,255,0))
        self.not_ready_text = self.small_font.render('NOT READY', True, (255,0,0))

        self.host_text = self.small_font.render('host', True, (255,0,0))
        self.escape_count = 0
        self.connected = False
        self.username = username

        p.mixer.init()
        self.sound_be = pygame.mixer.Sound('../assets/sfx/explosion-big.wav')
        self.sound_se = pygame.mixer.Sound('../assets/sfx/explosion-small.wav')
        self.sound_sp = pygame.mixer.Sound('../assets/sfx/splash.wav')
        self.sound_sh = pygame.mixer.Sound('../assets/sfx/splash-heavy.wav')
        self.sound_shot = pygame.mixer.Sound('../assets/sfx/shot.wav')
        self.sound_bump = pygame.mixer.Sound('../assets/sfx/bump.wav')
        self.sound_bb = pygame.mixer.Sound('../assets/sfx/building.wav')
        self.sound_ow = pygame.mixer.Sound('../assets/sfx/open_window.wav')
        self.sound_cw = pygame.mixer.Sound('../assets/sfx/close_window.wav')

        self.musics = {
            'steppingpebbles':'../assets/sfx/stepPebblesLoop.mp3',
            'village':'../assets/sfx/villageLoop.mp3',
            'barbarianraid':'../assets/sfx/War song.mp3'
            }

        self.paused = False

        with open('tips.txt') as tipfile:
            self.tips = [line.replace('\n', '') for line in tipfile]
        self.tip_frame = 0
        self.tip = 0


        self.achievement_pic = p.transform.scale(p.image.load('../assets/achievement.png'), (225, 50))
        
        self.achievement = [0, 'None']
        

        self.Connect((host, port))

        self.version = version
        connection.Send({'action':'version', 'version':self.version})

    def update(self):
        """
        Client update function. This is the function that runs over and
        over again, once every frame of the game.
        """
        self._connected = False
        connection.Pump()
        self.Pump()
        global running

        if answer == 'OK':
            self.connected = False
            connection.Close()
            p.quit()
            exit()

        global cursor
        old_cursor = p.mouse.get_cursor()
        if not find_cursor(cursor) == find_cursor(old_cursor):
            p.mouse.set_cursor(cursor)
        cursor = p.cursors.Cursor(pygame.SYSTEM_CURSOR_ARROW)

        if self.achievement[0]:
            self.achievement[0] -= 1
            if self.achievement[0] >= 90:
                x = (225/30) * (120-self.achievement[0])
            elif self.achievement[0] <= 30:
                x = (225/30) * self.achievement[0]
            else:
                x = 225
            self.toDraw.append((self.achievement_pic, (1000-x, 55)))
            text = self.small_font.render(self.achievement[1], True, (0,0,0))
            text_rect = text.get_rect(midtop = (1120-x, 85))
            self.toDraw.append([text, text_rect])

        if not self._connected:
            self.disconnected += 1
        else:
            self.disconnected = 0

        if self.escape_count:
            self.escape_count -= 1

        if self.disconnected == 50:
            p.mixer.music.stop()
            self.screen.blit(p.image.load('../assets/Disconnected.png'), (0, 0))
            pygame.display.flip()
            
            while True:                
                for e in p.event.get():
                    if e.type == QUIT:
                        p.quit()
                        sys.exit()
                        
                fps = self.clock.get_fps()
                self.clock.tick(30)
                pygame.display.set_caption("SuperShooterWar fps: " + str(fps))
        
        self.tip_frame += 1
        if self.tip_frame == 250:
            self.tip_frame = 0
            self.tip += 1
            if self.tip == len(self.tips):
                self.tip = 0
                random.shuffle(self.tips)
                if 'Always get a Miner\'s Guild and a Farmer\'s Guild first!' in self.tips:
                    self.tips.remove('Always get a Miner\'s Guild and a Farmer\'s Guild first!')

        
        # Set running to False if the player clicks the X
        global music_playing
        for event in pygame.event.get():
            if event.type == QUIT:
                def confirm_exit():
                    global answer
                    answer = confirm('The game is still going. Are you sure you want to quit?')
                thread = threading.Thread(target=confirm_exit)
                thread.start()
                
            if event.type == KEYUP and event.key == K_F1 and self.started:
                music_playing = not music_playing
                if music_playing:
                    p.mixer.music.pause()
                else:
                    p.mixer.music.play(-1, 0.0)
            if event.type == KEYUP and event.key == K_ESCAPE:
                connection.Send({'action':'escape'})
            if event.type == KEYUP and event.key == K_F2:
                num = len(os.listdir('../run/screenshots')) + 1
                name = 'sreenshot' + str(num)
                p.image.save(self.screen, '../run/screenshots/%s.png' % (name))
                alert('Screenshot saved as %s.png' % (name))
            if event.type == KEYUP and event.key == K_F10:
                hack()
                
            if event.type == KEYUP and event.key == K_ESCAPE:
                if not self.escape_count:
                    connection.Send({'action':'pause'})
                    self.escape_count = 30
            if event.type == MOUSEMOTION:
                self.mouseX = event.pos[0]
                self.mouseY = event.pos[1]
        _keys = p.key.get_pressed()
        keys = []
        for i in range(len(list(_keys))):
            keys.append(_keys[i])
        mouse = list(p.mouse.get_pos())
        mouse_click = p.mouse.get_pressed()
        mouse.append(mouse_click[0])
        mouse.append(mouse_click[1])
        mouse.append(mouse_click[2])

        connection.Send({'action': 'keys', 'keys': keys, 'mouse':mouse})

        
        
        
        _screen = p.Surface((1000, 650))
        for image in self.toDraw:
            _screen.blit(image[0], image[1])
        if self.paused:
            _screen.blit(self.paused_image, (0, 0))
        
        self.screen.blit(_screen, (0, 0))
        
        
            
        # Tell pygame to update the screen
        pygame.display.flip()

        fps = self.clock.get_fps()
        connection.Send({'action':'fps', 'fps':fps})
        self.clock.tick(self.fps_expected)
        pygame.display.set_caption("VillageWars " + self.version + " fps: " + str(fps))


    def ShutDown(self):
        """
        Client ShutDown function. Disconnects and closes the game.
        """
        self.connected = False
        connection.Close()
        
                
    #####################################
    ### Client-side Network functions ###
    #####################################
    """
    Each one of these "Network_" functions defines a command
    that the server will tell you (the client) to do.
    """
    def Network_hack_fail(self, data):
        print('Exception:', data['msg'])


    def Network_receive(self, data):
        #print(time.time() - data['timestamp'])
        self.diconnected = 0
        self._connected = True
        if time.time() - data['timestamp'] < 0.5:
            self.fps_expected = 30
            for i in data['data']:
                exec('self.Network_' + i['action'] + '(' + repr(i) + ')')
        else:
            self.fps_expected = 60
    

    def Network_fall(self, data):
        
        self.Setting = p.image.load('../assets/Setting.png')
        p.mixer.music.stop()
        p.mixer.music.load('../assets/sfx/villageLoop.mp3')
        global music_playing
        if music_playing:
            p.mixer.music.play(-1, 0.0)

    def Network_achievement(self, data):
        self.achievement = [120, data['type']]

    def Network_pause(self, data):
        self.paused = not self.paused

    def Network_flip(self, data):
        if not self.paused:           
            self.toDraw.clear()
            self.toDraw = self.toDrawNext[:]
            self.toDrawNext.clear()
        
        
    def Network_draw_setting(self, data):
        self.toDrawNext.append([self.Setting, data['coords']])

        
    def Network_draw_lobby_background(self, data):
        self.toDrawNext.append([self.back_pic, (0, 0)])

    def Network_draw_background(self, data):
        self.toDrawNext.append([self.ocean_pics[self.ocean_frame], (0, 0)])
        self.ocean_frame += 1
        self.ocean_frame %= 40

    def Network_preview(self, data):

        if not data['ArcheryTower?']:
            image = self.house
            rect = image.get_rect(center=(500, 185))
            cage = self.preview

            cage = p.transform.scale(cage, data['dimensions'])
            cage_rect = cage.get_rect(midtop=rect.midtop)
            self.toDrawNext.append([cage, cage_rect])
        else:
            cage = p.transform.scale(self.preview, (360, 360))
            cage_rect = cage.get_rect(center=(500, 85))
            self.toDrawNext.append([cage, cage_rect])


    def Network_draw_avatar(self, data):
        avatar_pic = self.players[data['skin']][0]
        avatar_rect = avatar_pic.get_rect(center=data['coords'])
        self.toDrawNext.append([avatar_pic, avatar_rect])

        if data['ready']:
            text_to_draw = self.ready_text
        else:
            text_to_draw = self.not_ready_text

        text_rect = text_to_draw.get_rect(midtop = avatar_rect.midbottom)
        text_rect.y += 4
        self.toDrawNext.append([text_to_draw, text_rect])

        num_text = self.medium_font.render(data['username'], True, data['color'])
        text_rect = num_text.get_rect(midbottom = avatar_rect.midtop)
        text_rect.y -= 4
        self.toDrawNext.append([num_text, text_rect])

        if data['host']:
            mouse = p.mouse.get_pos()
            click = p.mouse.get_pressed()

            image = self.x
            rect = image.get_rect(center=avatar_rect.topright)
            if rect.collidepoint(mouse):
                global cursor
                cursor = p.cursors.Cursor(pygame.SYSTEM_CURSOR_HAND)
                image = self.x_hov
                if click[0]:
                    connection.Send({'action':'hack', 'command':'kick(self, "%s")' % (data['username'])})

            self.toDrawNext.append((image, rect))
        

    def Network_victory(self, data):
        p.mixer.music.stop()
        victory = p.mixer.Sound('../assets/sfx/victory.mp3')
        victory.play()

    def Network_animation(self, data):
        image = self.animations[data['name']][data['frame']]
        rect = image.get_rect(center=data['coords'])
        self.toDrawNext.append((image, rect))

    def Network_congrats(self, data):
        screen = p.Surface((1000, 650))
        screen.fill((255,255,0))
        self.toDrawNext.append([screen, (0, 0)])
        if data['color'] == (255,255,0): data['color'] = (0,0,0)
        name = self.large_font.render('You won the game!', True, data['color'])
        text_rect = name.get_rect(midtop=(500, 30))
        self.toDrawNext.append([name, text_rect])

        name = self.medium_font.render('Kills: ' + str(data['kills']), True, (0,0,0))
        text_rect = name.get_rect(topleft=(400, 200))
        self.toDrawNext.append([name, text_rect])
        name = self.medium_font.render('Eliminations: ' + str(data['eliminations']), True, (0,0,0))
        text_rect = name.get_rect(topleft=(400, 320))
        self.toDrawNext.append([name, text_rect])
        name = self.medium_font.render('Buildings Destroyed: ' + str(data['destroyed']), True, (0,0,0))
        text_rect = name.get_rect(topleft=(400, 440))
        self.toDrawNext.append([name, text_rect])
        name = self.medium_font.render('Deaths: ' + str(data['deaths']), True, (255,0,0))
        text_rect = name.get_rect(topleft=(400, 560))
        self.toDrawNext.append([name, text_rect])

        keys = p.key.get_pressed()
        if keys[K_ESCAPE]:
            global running
            running = False

    def Network_end(self, data):
        p.mixer.music.stop()
        screen = p.Surface((1000, 650))
        screen.fill((128, 128, 128))
        self.toDrawNext.append([screen, (0, 0)])
        name = self.large_font.render(data['winner'] + ' won the game.', True, (0,0,0))
        text_rect = name.get_rect(midtop=(500, 30))
        self.toDrawNext.append([name, text_rect])

        name = self.medium_font.render('Kills: ' + str(data['kills']), True, (0,0,0))
        text_rect = name.get_rect(topleft=(400, 200))
        self.toDrawNext.append([name, text_rect])
        name = self.medium_font.render('Eliminations: ' + str(data['eliminations']), True, (0,0,0))
        text_rect = name.get_rect(topleft=(400, 320))
        self.toDrawNext.append([name, text_rect])
        name = self.medium_font.render('Buildings Destroyed: ' + str(data['destroyed']), True, (0,0,0))
        text_rect = name.get_rect(topleft=(400, 440))
        self.toDrawNext.append([name, text_rect])
        name = self.medium_font.render('Deaths: ' + str(data['deaths']), True, (255,0,0))
        text_rect = name.get_rect(topleft=(400, 560))
        self.toDrawNext.append([name, text_rect])
        
    def Network_ingame(self, data):
        self.toDrawNext.append([self.Setting, data['coords']])

    def Network_music_change(self, data):
        p.mixer.music.stop()
        p.mixer.music.load(self.musics[data['music']])
        global music_playing
        if music_playing:
            p.mixer.music.play(-1, 0.0)


    def Network_draw_barbarian(self, data):
        image = self.barbarians[data['type']][bool(data['hurt'])]
        rect = image.get_rect(center=data['coords'])
        image, rect = t.rotate(image, rect, data['angle'])

        if data['type'] == 'leader':
            banner = self.barbarian_banner
            banner_rect = banner.get_rect(bottomleft=rect.center)
            self.toDrawNext.append((banner, banner_rect))
        
        self.toDrawNext.append((image, rect))

        username_text = self.small_font.render('Barbarian ' + data['type'].title(), True, (0,0,0))
        text_rect = username_text.get_rect(midbottom = (rect.midtop[0], rect.midtop[1] - 15))
        self.toDrawNext.append([username_text, text_rect])


        if data['shield'] != None:
            image = self.shield
            rect = image.get_rect(center=data['coords'])
            image, rect = t.rotate(image, rect, data['shield'])
            self.toDrawNext.append((image, rect))
            
        

        
        health_bar = p.Surface((50, 10))
        health_bar.fill((255, 0, 0))
        health_rect = health_bar.get_rect(midbottom=rect.midtop)
        
        green_bar = p.Surface((int(data['health']/2), 10))
        green_bar.fill((0,255,0))
        health_bar.blit(green_bar, (0, 0))
        
        self.toDrawNext.append([health_bar, health_rect])


    def Network_draw_player(self, data):
        global cursor
        cursor = p.cursors.Cursor(pygame.SYSTEM_CURSOR_CROSSHAIR)
        image = self.players[data['skin']][data['hurt']]
        rect = image.get_rect(center=(data['coords']))
        image, rect = t.rotate(image, rect, data['angle'])
        self.toDrawNext.append([image, rect])


        
        username_text = self.small_font.render(data['username'], True, data['color'])
        text_rect = username_text.get_rect(midbottom = (rect.midtop[0], rect.midtop[1] - 15))
        self.toDrawNext.append([username_text, text_rect])

        
        health_bar = p.Surface((int(data['max_health'] / 2), 10))
        health_bar.fill((255, 0, 0))
        health_rect = health_bar.get_rect(midbottom=rect.midtop)
        
        green_bar = p.Surface((int(data['health']/2), 10))
        green_bar.fill((0,255,0))
        health_bar.blit(green_bar, (0, 0))
        
        self.toDrawNext.append([health_bar, health_rect])

    

        try:
            
            gold = self.medium_font.render('Gold: ' + str(data['gold']), True, (255,205,0))
            gold_rect = gold.get_rect(topright = (950,260))
            self.toDrawNext.append([gold, gold_rect])
            
            food = self.medium_font.render('Food: ' + str(data['food']), True, (5,255,5))
            food_rect = food.get_rect(topright = (950,300))
            self.toDrawNext.append([food, food_rect])
        except:
            pass

        if data['shield'] != None:
            image = self.shield
            rect = image.get_rect(center=data['coords'])
            image, rect = t.rotate(image, rect, data['shield'])
            self.toDrawNext.append((image, rect))


    def Network_draw_NPC(self, data):
        if data['image'] == 'farmer':
            image = self.farmer
            name = 'Farmer'
        elif data['image'] == 'miner':
            image = self.miner
            name = 'Miner'
        rect = image.get_rect(topleft=(data['coords']))
        image, rect = t.rotate(image, rect, data['angle'])
        self.toDrawNext.append([image, rect])


        
        username_text = self.small_font.render(name, True, data['color'])
        text_rect = username_text.get_rect(midbottom = (rect.midtop[0], rect.midtop[1] - 15))
        self.toDrawNext.append([username_text, text_rect])

        status_text = self.small_font.render(data['status'], True, (0,0,0))
        text_rect = status_text.get_rect(midbottom = (rect.midtop[0], rect.midtop[1] + 5))
        self.toDrawNext.append([status_text, text_rect])

    def Network_draw_InnPC(self, data):
        
        image = self.players[0][0]
        rect = image.get_rect(center=(data['coords']))
        image, rect = t.rotate(image, rect, data['angle'])
        self.toDrawNext.append([image, rect])
        
        username_text = self.small_font.render(data['type'], True, data['color'])
        text_rect = username_text.get_rect(midbottom = (rect.midtop[0], rect.midtop[1] - 15))
        self.toDrawNext.append([username_text, text_rect])

        health_bar = p.Surface((50, 10))
        health_bar.fill((255,0,0))
        health_rect = health_bar.get_rect(midbottom=rect.midtop)
        
        green_bar = p.Surface((round(data['health'] / 2), 10))
        green_bar.fill((0,255,0))
        health_bar.blit(green_bar, (0, 0))
        
        self.toDrawNext.append([health_bar, health_rect])
    def Network_draw_robot(self, data):
        if data['image'] == 'regular':
            image = self.robot
        elif data['image'] == 'hurt':
            image = self.robot_hurt
        
        rect = image.get_rect(center=(data['coords']))
        image, rect = t.rotate(image, rect, data['angle'])
        self.toDrawNext.append([image, rect])


        
        username_text = self.small_font.render(data['name'], True, data['color'])
        text_rect = username_text.get_rect(midbottom = (rect.midtop[0], rect.midtop[1] - 15))
        self.toDrawNext.append([username_text, text_rect])


        health_bar = p.Surface((50, 10))
        health_bar.fill((255,0,0))
        health_rect = health_bar.get_rect(midbottom=rect.midtop)
        
        green_bar = p.Surface((round(data['health'] / 2), 10))
        green_bar.fill((0,255,0))
        health_bar.blit(green_bar, (0, 0))
        
        self.toDrawNext.append([health_bar, health_rect])


    def Network_draw_farm(self, data):
        if data['state'] == 'good':
            image = self.food_image
        elif data['state'] == 'mine':
            image = self.food_mine
        else:
            image = self.food_used
        self.toDrawNext.append([image, data['coords']])

    def Network_draw_gold(self, data):
        image = None
        if data['state'] == 'good':
            image = self.gold_image
        elif data['state'] == 'mine':
            image = self.gold_mine
        if image != None:
            self.toDrawNext.append([image, data['coords']])

    def Network_draw_mine(self, data):
        image = self.mine
        if data['right'] == True:
            image = p.transform.flip(image, True, False)
        self.toDrawNext.append([image, data['coords']])
    

    def Network_draw_balloon(self, data):
        balloon_type = data.get('type', 'normal')
        image = self.balloon
        if balloon_type == 'bolt':
            image = self.bolt
        if balloon_type == 'op':
            image = self.op_balloon
        if balloon_type == 'speedy':
            image = self.speedy_balloon
        if balloon_type == 'speedy+op':
            image = self.speedy_plus_op
        rect = image.get_rect(center=(data['coords']))
        image, rect = t.rotate(image, rect, data['angle'])
        self.toDrawNext.append([image, rect])

        
    def Network_draw_building(self, data):
        
        if data['image'] == 'house':
            if data['state'] == 'alive':
                image = self.house
            else:
                image = self.house_burnt
        rect = image.get_rect(center=data['coords'])
        

        color = data['color']
        if color == (255,0,0):
            cage = self.red
        elif color == (0,255,0):
            cage = self.green
        elif color == (0,0,255):
            cage = self.blue
        elif color == (255,255,0):
            cage = self.yellow

        cage = p.transform.scale(cage, data['dimensions'])
        cage_rect = cage.get_rect(midtop=rect.midtop)
        self.toDrawNext.append([cage, cage_rect])

        self.toDrawNext.append([image, rect])

        if data['state'] == 'alive':
            person = self.building_person
            person_rect = image.get_rect(midtop=(rect.midbottom[0], rect.midbottom[1] + 50))
            person, person_rect = t.rotate(person, person_rect, data['angle'])

            self.toDrawNext.append([person, person_rect])

            name = self.small_font.render(data['type'], True, data['color'])
            text_rect = name.get_rect(midbottom = (person_rect.midtop[0], person_rect.midtop[1] - 30))
            self.toDrawNext.append([name, text_rect])

            name = self.small_font.render('(Level ' + str(data['level']) + ')', True, data['color'])
            text_rect = name.get_rect(midbottom = (person_rect.midtop[0], person_rect.midtop[1] - 15))
            self.toDrawNext.append([name, text_rect])

            health_bar = p.Surface((int(data['max_health'] / 2), 10))
            health_bar.fill((255, 0, 0))
            health_rect = health_bar.get_rect(midbottom=person_rect.midtop)
            
            green_bar = p.Surface((int(data['health']/2), 10))
            green_bar.fill((0,255,0))
            health_bar.blit(green_bar, (0, 0))
            
            self.toDrawNext.append([health_bar, health_rect])

    def Network_archery_tower(self, data):
        

        color = data['color']
        if color == (255,0,0):
            cage = self.red
        elif color == (0,255,0):
            cage = self.green
        elif color == (0,0,255):
            cage = self.blue
        elif color == (255,255,0):
            cage = self.yellow

        cage = p.transform.scale(cage, (360, 360))
        cage_rect = cage.get_rect(center=data['coords'])
        self.toDrawNext.append([cage, cage_rect])

        
        
        person = self.archer
        person_rect = person.get_rect(midtop=(data['coords']))
        person, person_rect = t.rotate(person, person_rect, data['angle'])
        if data['state'] == 'alive':
            self.toDrawNext.append([person, person_rect])

        if data['state'] == 'alive':
            name = self.small_font.render('Archery Tower', True, data['color'])
        else:
            name = self.small_font.render('Archery Tower (Broken)', True, data['color'])
        text_rect = name.get_rect(midbottom = (person_rect.midtop[0], person_rect.midtop[1] - 15))
        self.toDrawNext.append([name, text_rect])

        if data['state'] == 'alive':
            health_bar = p.Surface((150, 10))
            health_rect = health_bar.get_rect(midbottom=person_rect.midtop)
            health_bar.fill((255,0,0))
            
            green_bar = p.Surface((int(data['health']/2), 10))
            green_bar.fill((0,255,0))
            health_bar.blit(green_bar, (0, 0))
            
            self.toDrawNext.append([health_bar, health_rect])

    def Network_sound(self, data):
        if data['sound'] == 'TNT':
            self.sound_be.play()
        if data['sound'] == 'shot':
            self.sound_shot.play()
        if data['sound'] == 'die':
            self.sound_se.play()
        if data['sound'] == 'splash':
            self.sound_sp.play()
        if data['sound'] == 'opsplash':
            self.sound_sh.play()
        if data['sound'] == 'bump':
            self.sound_bump.play()
        if data['sound'] == 'building':
            self.sound_bb.play()
        if data['sound'] == 'ow':
            self.sound_ow.play()
        if data['sound'] == 'cw':
            self.sound_cw.play()


    def Network_num_buildings(self, data):
        y = 500
        x = 990
        for user in data['users']:
            username = self.small_font.render(user['name'] + ' has ' + str(user['bs']) + ' buildings.', True, user['color'])
            text_rect = username.get_rect(topright=(x, y+user['num']*30))
            self.toDrawNext.append([username, text_rect])
    

    def Network_chat(self, data):

        name = self.medium_font.render(data['message'], True, data['color'])
        text_rect = name.get_rect(topleft=(20, 20))
        self.toDrawNext.append([name, text_rect])

    def Network_startgame(self, data):
        p.mixer.music.stop()
        p.mixer.music.load('../assets/sfx/stepPebblesLoop.mp3')
        self.started = True
        global music_playing
        if music_playing:
            p.mixer.music.play(-1, 0.00)

    def Network_text(self, data):

        if data['text']:
            name = self.medium_font.render(data['text'], True, (0,0,255))
        else:
            name = self.medium_font.render('TIP: ' + self.tips[self.tip], True, (255,205,0))            
        text_rect = name.get_rect(midbottom=(500, 500))
        self.toDrawNext.append([name, text_rect])
    
    def Network_crates(self, data):
        self.toDrawNext.append([p.transform.scale(self.crate, (30, 30)), (900, 5)])
        
        text = self.medium_font.render('x ' + str(data['crates']), True, (0,0,0))
        text_rect = text.get_rect(midleft=(940, 20))
        self.toDrawNext.append([text, text_rect])

    def Network_spiky_bushes(self, data):
        self.toDrawNext.append([p.transform.scale(self.spiky_bush, (30, 30)), (900, 60)])
        
        text = self.medium_font.render('x ' + str(data['num']), True, (0,0,0))
        text_rect = text.get_rect(midleft=(940, 75))
        self.toDrawNext.append([text, text_rect])

    def Network_time(self, data):
        text = self.medium_font.render(data['time'], True, (0,0,0))
        text_rect = text.get_rect(midbottom=(500, 640))
        self.toDrawNext.append([text, text_rect])

    def Network_draw_obstacle(self, data):
        if data['image'] == 'tree':
            image = self.tree
        if data['image'] == 'sappling':
            image = self.sappling
        if data['image'] == 'crate':
            image = self.crate
        if data['image'] == 'vine':
            image = self.vine
        if data['image'] == 'gate':
            if data['rotated?']:
                image = p.transform.rotate(self.gate, 90)
            else:
                image = self.gate
        if data['image'] == 'spiky bush':
            image = self.spiky_bush
        if data['image'] == 'TNT':
            image = self.tnt
        rect = image.get_rect(center=data['coords'])
        self.toDrawNext.append([image, rect])

        if data['image'] != 'crate' and data['image'] != 'gate' and data['image'] != 'TNT' and data['image'] != 'spiky bush' and data['image'] != 'vine':
            health_bar = p.Surface((int(data['max_health']/2), 12))
            health_bar.fill((255, 0, 0))
            health_rect = health_bar.get_rect(midbottom=rect.midtop)
        
            green_bar = p.Surface((int(data['health']/2), 12))
        
            green_bar.fill((0,255,0))
            health_bar.blit(green_bar, (0, 0))
        
            self.toDrawNext.append([health_bar, health_rect])

    def Network_meal(self, data):
        if data['food?']:
            image = self.meals[data['type']]
        else:
            image = self.plate

        mouse = p.mouse.get_pos()

        rect = image.get_rect(center=(890, 130))
        if rect.collidepoint(mouse) and image != self.plate:
            image = p.transform.scale(image, (110, 110))
            rect = image.get_rect(center=(890, 130))
            pressed = p.mouse.get_pressed()
            if pressed[0]:
                connection.Send({'action':'eat'})

        self.toDrawNext.append([image, rect])


    def Network_draw_window(self, data):
        grey = p.Surface((610, 618))
        large_rect = grey.get_rect(topleft=(195, 4))
        grey.fill((128,128,128))
        self.toDrawNext.append([grey, large_rect])

        if large_rect.collidepoint(p.mouse.get_pos()):
            global cursor
            cursor = p.cursors.Cursor(pygame.SYSTEM_CURSOR_ARROW)

        dark_grey = p.Surface((600, 606))
        rect = dark_grey.get_rect(topleft=(200, 10))
        dark_grey.fill((95,95,95))
        self.toDrawNext.append([dark_grey, rect])

        x = 200
        y = 200

        info = self.small_font.render(data['window']['info'][0], True, (0,0,0))
        text_rect = info.get_rect(topleft=(210,30))
        self.toDrawNext.append([info, text_rect])
        info = self.small_font.render(data['window']['info'][1], True, (0,0,0))
        text_rect = info.get_rect(topleft=(210,55))
        self.toDrawNext.append([info, text_rect])

        info = self.small_font.render('Owner:', True, (255,0,255))
        text_rect = info.get_rect(topleft=(210,90))
        self.toDrawNext.append([info, text_rect])
        info = self.small_font.render('Health:', True, (255,0,255))
        text_rect = info.get_rect(topleft=(210,110))
        self.toDrawNext.append([info, text_rect])
        info = self.small_font.render('Level:', True, (255,0,255))
        text_rect = info.get_rect(topleft=(210,130))
        self.toDrawNext.append([info, text_rect])
        info = self.small_font.render(data['window']['4th_info'][0] + ':', True, (255,0,255))
        big_text_rect = info.get_rect(topleft=(210,150))
        self.toDrawNext.append([info, big_text_rect])
        right = big_text_rect.topright[0] + 30

        info = self.small_font.render(data['window']['owner'], True, data['window']['color'])
        text_rect = info.get_rect(topleft=(right,90))
        self.toDrawNext.append([info, text_rect])

        info = self.small_font.render(str(data['window']['health'][0]) + '/' + str(data['window']['health'][1]), True, (0,0,0))
        text_rect = info.get_rect(topleft=(right,110))
        self.toDrawNext.append([info, text_rect])

        info = self.small_font.render(str(data['window']['level']), True, (0,0,0))
        text_rect = info.get_rect(topleft=(right,130))
        self.toDrawNext.append([info, text_rect])

        info = self.small_font.render(data['window']['4th_info'][1], True, (0,0,0))
        text_rect = info.get_rect(topleft=(right,150))
        self.toDrawNext.append([info, text_rect])

        for Y in range(len(data['window']['options'])):
            black = p.Surface((600, 2))
            black_rect = black.get_rect(topleft=(200,y+Y*40))
            grey = p.Surface((600, 44))
            rect = grey.get_rect(topleft=(200, y+Y*40))
            grey.fill((155,155,155))
            mouse = p.mouse.get_pos()
            if rect.collidepoint(mouse):
                grey.fill((250, 250, 255))
            self.toDrawNext.append([grey, rect])
            self.toDrawNext.append([black, black_rect])

            name = self.medium_font.render(data['window']['options'][Y], True, (0,0,0))
            text_rect = name.get_rect(midtop=(500, y+Y*40+5))
            self.toDrawNext.append([name, text_rect])


    def Network_draw_innpc_window(self, data):
        grey = p.Surface((610, 618))
        large_rect = grey.get_rect(topleft=(195, 4))
        grey.fill((128,128,128))
        self.toDrawNext.append([grey, large_rect])

        dark_grey = p.Surface((600, 606))
        rect = dark_grey.get_rect(topleft=(200, 10))
        dark_grey.fill((95,95,95))
        self.toDrawNext.append([dark_grey, rect])

        x = 210
        y = 30

        for Y, thing in enumerate(data['window']['info']):
            info = self.small_font.render(thing, True, (0,0,0))
            text_rect = info.get_rect(topleft=(x, y + Y*15))
            self.toDrawNext.append([info, text_rect])

        x = 200
        y = 200
        for Y in range(len(data['window']['options'])):
            black = p.Surface((600, 2))
            black_rect = black.get_rect(topleft=(200,y+Y*40))
            grey = p.Surface((600, 44))
            rect = grey.get_rect(topleft=(200, y+Y*40))
            grey.fill((155,155,155))
            mouse = p.mouse.get_pos()
            if rect.collidepoint(mouse):
                grey.fill((250, 250, 255))
            self.toDrawNext.append([grey, rect])
            self.toDrawNext.append([black, black_rect])

            name = self.medium_font.render(data['window']['options'][Y], True, (0,0,0))
            text_rect = name.get_rect(midtop=(500, y+Y*40+5))
            self.toDrawNext.append([name, text_rect])

    def Network_WARNING_outdated_client(self, data):
        info = self.medium_font.render('WARNING: Outdated Client', True, (255,0,0))
        text_rect = info.get_rect(midbottom=(400,600))

        box = p.Surface((text_rect.width + 10, 85))
        box.fill((255,255,255))
        rect = box.get_rect(midbottom=(400, 650))
        self.toDrawNext.append([box, rect])

        self.toDrawNext.append([info, text_rect])
        info = self.medium_font.render('Server: %s - Client: %s' % (data['version'], self.version), True, (255,0,0))
        text_rect = info.get_rect(midbottom=(400,635))
        self.toDrawNext.append([info, text_rect])

    


    def Network_WARNING_outdated_server(self, data):
        info = self.medium_font.render('WARNING: Outdated Server', True, (255,0,0))
        text_rect = info.get_rect(midbottom=(400,600))

        box = p.Surface((text_rect.width + 10, 85))
        box.fill((255,255,255))
        rect = box.get_rect(midbottom=(400, 650))
        self.toDrawNext.append([box, rect])

        self.toDrawNext.append([info, text_rect])
        info = self.medium_font.render('Server: %s - Client: %s' % (data['version'], self.version), True, (255,0,0))
        text_rect = info.get_rect(midbottom=(400,635))
        self.toDrawNext.append([info, text_rect])

        
        

    def Network_display_host(self, data):
        self.toDrawNext.append([self.host_text, (480, 20)])

        if data['enough?']:
            if self.startgame_rect.collidepoint((self.mouseX, self.mouseY)):
                image = self.startgame_icon_over
                if p.mouse.get_pressed()[0]:
                    print(p.mouse.get_pressed()[0])
                    connection.Send({'action':'all_ready'})
            else:
                image = self.startgame_icon
        else:
            image = self.startgame_icon_locked

        self.toDrawNext.append([image, self.startgame_rect])


        

    def Network_connected(self, data):
        """
        Network_connected runs when you successfully connect to the server
        """
        
        connection.Send({'action':'init', 'username':self.username, 'status':'JG', 'color':self.color, 'skin':self.skin})
        self.connected = True
        print("Joined game")

        p.mixer.music.stop()
        
    
    def Network_error(self, data):
        """
        Network_error runs when there is a server error
        """
        print('error:', data['error'][1])
        self.ShutDown()
    
    def Network_disconnected(self, data):
        """
        Network_disconnected runs when you disconnect from the server
        """
        self.ShutDown()

    def Network_kicked(self, data):
        self.ShutDown()
        p.quit()
        exit()






    
def main(screen, clock, username, version, userInfo, ip, port=5555, musicPlaying=True):
    global running, cursor, music_playing
    running = True
    cursor = p.cursors.Cursor(p.SYSTEM_CURSOR_ARROW)
    music_playing = musicPlaying
    print(ip)
    cursor = p.cursors.Cursor(pygame.SYSTEM_CURSOR_ARROW)
    client = MyGameClient(ip, port, screen, clock, username, version, userInfo['color'], userInfo['skin'])
    while running:
        client.update()
    client.ShutDown()
    p.mixer.music.stop()
    return music_playing


if __name__ == '__main__':
    import VillageWarsClient
    VillageWarsClient.main()
                   




















import pygame
import toolbox as t
from animations import *
import random as r


class Balloon(pygame.sprite.Sprite):
    def __init__(self, server, owner):
        pygame.sprite.Sprite.__init__(self, self.gp)
        self.server = server
        self.owner = owner
        self._owner = owner.channel.username
        self.speed = owner.balloon_speed
        self.x = owner.x
        self.y = owner.y
        self.angle = owner.angle
        self.damage = owner.attack
        self.knockback = owner.knockback

        self.image = pygame.image.load('../assets/balloon.png')
        self.type = ('normal' if self.damage < 20 else 'op')
        if self.owner.shot_speed < 8:
            self.type = ('speedy' if self.damage < 20 else 'speedy+op')


    def update(self):
        obstacles = self.server.obs
        
        self.move()

        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'draw_balloon',
                    'coords':(p.character.get_x(self), p.character.get_y(self)),
                    'angle':self.angle,
                    'type':self.type})

        if self.x > 7000 or self.x < -1000 or self.y > 4400 or self.y < -500:
            self.kill()

        self.rect = self.image.get_rect(center=(self.x, self.y))

        for item in obstacles:
            if self.rect.colliderect(item.innerrect) and not (self.__class__ == Arrow  and item.__class__.__name__ == 'ArcheryTower') and not (self.__class__ == Arrow  and item.owner == self.owner):
                try:
                    item.getHurt(self.damage, self.owner)
                except:
                    item.getHurt(self.damage, self.owner, 0)
                self.pop()
        for item in self.server.bushes:
            if self.rect.colliderect(item.innerrect) and not (self.__class__ == Arrow  and item.owner == self.owner):
                item.getHurt(self.damage, self.owner)
                self.pop()
                
                
            elif item.isBuilding():
                if self.rect.colliderect(item.rect) and item.owner != self.owner.channel:
                    item.getHurt(self.damage, self.owner)
                    self.pop()

                    
        for p in self.server.players:
            if not p.pending:
                if self.rect.colliderect(p.character.rect) and p != self.owner.channel and p.character.dead == False:
                    if self.__class__ == Arrow:
                        result = p.character.getHurt(self.damage, self._owner, self.angle, self.knockback, '<Victim> was shot by <Attacker>')
                    else:
                        result = p.character.getHurt(self.damage, self.owner, self.angle, self.knockback)
                    if result == 'repelled':
                        bb = type(self)(self.server, p.character)
                        bb.damage = self.owner.attack
                        bb.knockback = self.owner.knockback
                        bb.angle = self.angle + 180 + r.randint(-20, 20)
                        self.speed = self.owner.balloon_speed
                        self.kill()
                        return
                    else: 
                        self.pop()

                    
        for npc in self.server.NPCs:
            if npc.__class__.__name__ == 'Robot':
                if self.rect.colliderect(npc.rect) and not (self.owner == npc.player) and not npc.dead:
                    npc.getHurt(self.damage, self.angle, self.knockback)
                    self.pop()

        
        for npc in self.server.event.NPCs:
            if self.rect.colliderect(npc.rect):
                result = npc.getHurt(self.owner, self.damage, self.angle, self.knockback)
                if result and result[0] == 'repelled' and self.server.event.__class__.__name__ == 'BarbarianRaid':
                    BounceBalloon(self, result[1])
                    self.kill()
                else:
                    self.pop()
                    

    def move(self):

        x, y = t.getDir(self.angle, self.speed)
        self.x += x
        self.y += y

    def pop(self):
        for p in self.server.players:
            if not p.pending:
                screen = pygame.Rect(0, 0, 1000, 650)
                screen.center = p.character.rect.center
                if screen.colliderect(self.rect):
                    if self.type == 'normal' or self.type == 'speedy':
                        p.to_send.append({'action':'sound', 'sound':'splash'})
                    elif self.type == 'op' or self.type == 'speedy+op':
                        p.to_send.append({'action':'sound', 'sound':'opsplash'})
                    
        Splash(self)
        self.kill()



class Arrow(Balloon):
    def __init__(self, tower):
        Balloon.__init__(self, tower.server, tower.owner)
        self.x = tower.innerrect.center[0]
        self.y = tower.innerrect.center[1]
        self.speed = tower.balloon_speed
        self.angle = tower.angle
        self.damage = tower.attack
        self.knockback = tower.knockback
        self._owner = tower.owner.channel.username + "'s Archery Tower"
        self.type = 'normal'

class Bolt(pygame.sprite.Sprite):
    def __init__(self, archer):
        pygame.sprite.Sprite.__init__(self, Balloon.gp)
        if archer.__class__.__name__ == 'Character':
            self.server = archer.channel.server
            self.speed = archer.balloon_speed
            self.x = archer.x
            self.y = archer.y
            self.angle = archer.angle
            self.damage = 10
            self.knockback = 50
            self.owner = archer
            self.archer = None
        else:
            self.server = archer.event.server
            self.speed = 17
            self.x = archer.x
            self.y = archer.y
            self.angle = archer.angle
            self.damage = 10
            self.knockback = 50
            self.owner = 'A Barbarian Archer'
            self.archer = archer

        self.image = pygame.image.load('../assets/balloon.png')


    def update(self):
        obstacles = self.server.obs
        
        self.move()

        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'draw_balloon',
                    'coords':(p.character.get_x(self), p.character.get_y(self)),
                    'angle':self.angle,
                    'type':'bolt'})

        if self.x > 7000 or self.x < -1000 or self.y > 4400 or self.y < -500:
            self.kill()

        self.rect = self.image.get_rect(center=(self.x, self.y))

        for item in obstacles:
            if self.rect.colliderect(item.innerrect):
                item.getHurt(self.damage, self.owner)
                self.pop()
                
                
            elif item.isBuilding() and not (not self.archer and item.owner != self.owner.channel):
                if self.rect.colliderect(item.rect):
                    item.getHurt(self.damage, self.owner)
                    self.pop()

                    
        for p in self.server.players:
            if not (not self.archer and p.character == self.owner):
                if not p.pending:
                    if self.rect.colliderect(p.character.rect) and p.character.dead == False:
                        if self.archer:
                            p.character.getHurt(self.damage, 'A Barbarian Archer', self.angle, self.knockback, msg='<Attacker> shot <Victim> and stole their gold and food.')
                            if p.character.dead:
                                self.archer.gold += p.character.gold
                                self.archer.food += p.character.food
                                p.character.gold = 0
                                p.character.food = 0

                            self.pop()
                        else:
                            p.character.getHurt(self.damage, self.owner, self.angle, self.knockback, msg='<Attacker> shot <Victim> with a barbarian crossbow.')

                            self.pop()

                    
        for npc in self.server.NPCs:
            if npc.__class__.__name__ == 'Robot' and not (self.archer and npc.player == self.owner):
                if self.rect.colliderect(npc.rect) and not npc.dead:
                    npc.getHurt(self.damage, self.angle, self.knockback)
                    self.pop()

        if self.server.event.__class__.__name__ != 'BarbarianRaid' or not self.archer:
            for npc in self.server.event.NPCs:
                if self.rect.colliderect(npc.rect):
                    npc.getHurt(self.owner, self.damage, self.angle, self.knockback)

    def move(self):

        x, y = t.getDir(self.angle, self.speed)
        self.x += x
        self.y += y

    def pop(self):
        self.kill()

class BounceBalloon(pygame.sprite.Sprite):
    def __init__(self, old_self, angle):
        pygame.sprite.Sprite.__init__(self, Balloon.gp)
        self.server = old_self.server
        self.owner = 'A Barbarian'
        self.speed = old_self.speed
        self.x = old_self.x
        self.y = old_self.y
        self.angle = angle
        self.damage = old_self.damage
        self.knockback = old_self.knockback
        self.mock_owner = old_self.owner
        self.type = old_self.type

        self.image = pygame.image.load('../assets/balloon.png')


    def update(self):
        obstacles = self.server.obs
        
        self.move()

        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'draw_balloon',
                    'coords':(p.character.get_x(self), p.character.get_y(self)),
                    'angle':self.angle,
                    'type':self.type})

        if self.x > 7000 or self.x < -1000 or self.y > 4400 or self.y < -500:
            self.kill()

        self.rect = self.image.get_rect(center=(self.x, self.y))

        for item in obstacles:
            if self.rect.colliderect(item.innerrect):
                item.getHurt(self.damage, self.owner)
                self.pop()
                
                
            elif item.isBuilding():
                if self.rect.colliderect(item.rect):
                    item.getHurt(self.damage, self.owner)
                    self.pop()

                    
        for p in self.server.players:
            if not p.pending:
                if self.rect.colliderect(p.character.rect) and p.character.dead == False:
                    p.character.getHurt(self.damage, self.owner, self.angle, self.knockback)
                    self.pop()

                    
        for npc in self.server.NPCs:
            if npc.__class__.__name__ == 'Robot':
                if self.rect.colliderect(npc.rect) and not npc.dead:
                    npc.getHurt(self.damage, self.angle, self.knockback)
                    self.pop()

    def move(self):

        x, y = t.getDir(self.angle, self.speed)
        self.x += x
        self.y += y

    def pop(self):
        for p in self.server.players:
            if not p.pending:
                screen = pygame.Rect(0, 0, 1000, 650)
                screen.center = p.character.rect.center
                if screen.colliderect(self.rect):
                    p.to_send.append({'action':'sound', 'sound':'splash'})
                    
        Splash(self)
        self.kill()









        




class Background():

    def __init__(self, server):
        self.server = server
        self.x = 0
        self.y = 0
        

    def update(self):

        for p in self.server.players:
            if not p.pending:
                p.to_send.append({'action':'draw_setting', 'coords':(p.character.get_x(self), p.character.get_y(self))})



from flask import Flask
import shelve
import pyperclip
app = Flask(__name__)

@app.route('/')
def html():
    users = shelve.open('database/data')
    final = '''<!DOCTYPEhtml>
<html>
    <head>
        <title>Scores - VillageWars</title>
    </head>
    <body>
        <script>

// TODO: Add this code.

        </script>
        <h1>VillageWars</h1>
        <h2 style="color:blue">Scores</h2>
'''

    for username in users.keys():
        final += '<h3>' + username + '''</h3>
<h4>Victories: %s</h4>
<h4>Games finished: %s</h4>
<h4>Total Kills: %s</h4>
<h4>Color: %s</h4>
<br>''' % (users[username]['victories'], users[username]['games finished'], users[username]['kills'], users[username]['color'])
    final += '''
    </body>
</html>'''
    return final

app.run()
pyperclip.copy('http://127.0.0.1:5000/')



import pyperclip
import pygame as p
from pygame.locals import *

class Typing:
    def __init__(self, character='|'):
        self.typing = character
        self.character = character
        self.shifting = False
        self.ctrling = False

    @property
    def result(self):
        return self.typing[:self.typing.find(self.character)] + self.typing[self.typing.find(self.character) + 1:]


    def shift(self):
        self.shifting = not self.shifting

    def ctrl(self):
        self.ctrling = not self.ctrling

    def type(self, event):
        if event.key == K_a:
            self.add('a')
        if event.key == K_b:
            self.add('b')
        if event.key == K_c:
            self.add('c')
        if event.key == K_d:
            self.add('d')
        if event.key == K_e:
            self.add('e')
        if event.key == K_f:
            self.add('f')
        if event.key == K_g:
            self.add('g')
        if event.key == K_h:
            self.add('h')
        if event.key == K_i:
            self.add('i')
        if event.key == K_j:
            self.add('j')
        if event.key == K_k:
            self.add('k')
        if event.key == K_l:
            self.add('l')
        if event.key == K_m:
            self.add('m')
        if event.key == K_n:
            self.add('n')
        if event.key == K_o:
            self.add('o')
        if event.key == K_p:
            self.add('p')
        if event.key == K_q:
            self.add('q')
        if event.key == K_r:
            self.add('r')
        if event.key == K_s:
            self.add('s')
        if event.key == K_t:
            self.add('t')
        if event.key == K_u:
            self.add('u')
        if event.key == K_v and not self.ctrling:
            self.add('v')
        if event.key == K_w:
            self.add('w')
        if event.key == K_x:
            self.add('x')
        if event.key == K_y:
            self.add('y')
        if event.key == K_z:
            self.add('z')
        if event.key == K_1:
            self.add('1')
        if event.key == K_2:
            self.add('2')
        if event.key == K_3:
            self.add('3')
        if event.key == K_4:
            self.add('4')
        if event.key == K_5:
            self.add('5')
        if event.key == K_6:
            self.add('6')
        if event.key == K_7:
            self.add('7')
        if event.key == K_8:
            self.add('8')
        if event.key == K_9:
            self.add('9')
        if event.key == K_0:
            self.add('0')
        if event.key == K_PERIOD:
            self.add('.')
        if event.key == K_SLASH:
            self.add('/')

        if event.key == K_SPACE:
            self.add(' ')

        if event.key == K_BACKSPACE:
            self.typing = self.typing[:-2] + self.character

        if event.key == K_v and self.ctrling:
            self.typing = self.typing[:self.typing.find(self.character)] + pyperclip.paste() + self.typing[self.typing.find(self.character):]
        
    def add(self, letter):
        if self.shifting:
            letter = letter.upper()

        self.typing = self.typing[:self.typing.find(self.character)] + letter + self.typing[self.typing.find(self.character):]

    def text(self):
        return self.typing
